"""
Systemd File Permissions parsers
================================

Parsers included in this module are:

LsEtcSystemd - command ``ls -lanRL /etc/systemd``
-------------------------------------------------

LsUsrLibSystemd - command ``ls -lanRL /usr/lib/systemd``
--------------------------------------------------------

"""
from insights.core import CommandParser, FileListing
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.ls_etc_systemd)
class LsEtcSystemd(CommandParser, FileListing):
    """
    Class for parsing ``ls -lanRL /etc/systemd`` command.

    The ``ls -lanRL /etc/systemd`` command provides information for the listing of the ``/etc/systemd`` directory.

    See the ``FileListing`` class for a more complete description of the available features of the class.

    Sample output of ``ls -lanRL /etc/systemd`` command is::

        /etc/systemd:
        total 48
        drwxr-xr-x   10 501  20   320 Feb 23  2021 .
        drwxr-xr-x  129 501  20  4128 Nov  7 15:01 ..
        -rw-r--r--    1 501  20   720 Jan 16  2021 bootchart.conf
        -rw-r--r--    1 501  20   615 Jan 16  2021 coredump.conf
        drwxr-xr-x    2 501  20    64 Jan 16  2021 user
        -rw-r--r--    1 501  20  1127 Jan 16  2021 user.conf

        /etc/systemd/user:
        total 0
        drwxr-xr-x   2 501  20   64 Jan 16  2021 .
        drwxr-xr-x  10 501  20  320 Feb 23  2021 ..

    Examples:
        >>> type(etc_systemd)
        <class 'insights.parsers.ls_systemd.LsEtcSystemd'>
        >>> '/etc/systemd' in etc_systemd
        True
        >>> '/etc/systemd/system' in etc_systemd
        False
        >>> etc_systemd.files_of('/etc/systemd')
        ['bootchart.conf', 'coredump.conf', 'user.conf']
        >>> etc_systemd.dirs_of('/etc/systemd')
        ['.', '..', 'user']
        >>> etc_systemd.dir_contains('/etc/systemd', 'coredump.conf')
        True
        >>> etc_systemd_dict = etc_systemd.listing_of('/etc/systemd')
        >>> etc_systemd_dict['coredump.conf']['perms']
        'rw-r--r--'
        >>> etc_systemd_dict['user']['perms']
        'rwxr-xr-x'

"""
    pass


@parser(Specs.ls_usr_lib_systemd)
class LsUsrLibSystemd(CommandParser, FileListing):
    """
    Class for parsing ``ls -lanRL /usr/lib/systemd`` command.

    The ``ls -lanRL /usr/lib/systemd`` command provides information for the listing of the ``/usr/lib/systemd`` directory.

    See the ``FileListing`` class for a more complete description of the available features of the class.

    Sample output of ``ls -lanRL /usr/lib/systemd`` command is::

        /usr/lib/systemd:
        total 0
        drwxr-xr-x    4 501  20    128 Feb 23  2021 .
        dr-xr-xr-x    8 501  20    256 Oct 13 01:34 ..
        drwxr-xr-x  404 501  20  12928 Oct 13 02:50 system
        drwxr-xr-x   15 501  20    480 Feb 23  2021 user

        /usr/lib/systemd/user:
        total 104
        drwxr-xr-x  15 501  20  480 Feb 23  2021 .
        drwxr-xr-x   4 501  20  128 Feb 23  2021 ..
        -rw-r--r--   1 501  20  457 Jan 16  2021 basic.target
        -rw-r--r--   1 501  20  379 Jan 16  2021 bluetooth.target
        -rw-r--r--   1 501  20  414 Jan 16  2021 default.target
        -rw-r--r--   1 501  20  499 Jan 16  2021 exit.target
        -rw-r--r--   1 501  20  147 May 31  2018 glib-pacrunner.service
        -rw-r--r--   1 501  20  354 Jan 16  2021 paths.target
        -rw-r--r--   1 501  20  377 Jan 16  2021 printer.target
        -rw-r--r--   1 501  20  402 Jan 16  2021 shutdown.target
        -rw-r--r--   1 501  20  380 Jan 16  2021 smartcard.target
        -rw-r--r--   1 501  20  356 Jan 16  2021 sockets.target
        -rw-r--r--   1 501  20  380 Jan 16  2021 sound.target
        -rw-r--r--   1 501  20  501 Jan 16  2021 systemd-exit.service
        -rw-r--r--   1 501  20  405 Jan 16  2021 timers.target

    Examples:
        >>> type(usr_lib_systemd)
        <class 'insights.parsers.ls_systemd.LsUsrLibSystemd'>
        >>> '/usr/lib/systemd' in usr_lib_systemd
        True
        >>> '/usr/lib/systemd/system' in usr_lib_systemd
        False
        >>> usr_lib_systemd.files_of('/usr/lib/systemd/user')
        ['basic.target', 'bluetooth.target', 'default.target', 'exit.target', 'glib-pacrunner.service', 'paths.target', 'printer.target', 'shutdown.target', 'smartcard.target', 'sockets.target', 'sound.target', 'systemd-exit.service', 'timers.target']
        >>> usr_lib_systemd.dirs_of('/usr/lib/systemd')
        ['.', '..', 'system', 'user']
        >>> usr_lib_systemd.dir_contains('/usr/lib/systemd/user', 'basic.target')
        True
        >>> usr_lib_systemd_dict = usr_lib_systemd.listing_of('/usr/lib/systemd/user')
        >>> usr_lib_systemd_dict['basic.target']['perms']
        'rw-r--r--'
        >>> usr_lib_systemd_dir_dict = usr_lib_systemd.listing_of('/usr/lib/systemd')
        >>> usr_lib_systemd_dir_dict['user']['perms']
        'rwxr-xr-x'

    """
    pass



"""
getenforce - command ``/usr/sbin/getenforce``
=============================================

This very simple parser returns the output of the ``getenforce`` command.

Examples:

    >>> enforce = shared[getenforcevalue]
    >>> enforce['status']
    'Enforcing'
"""

from .. import parser
from insights.specs import Specs


@parser(Specs.getenforce)
def getenforcevalue(context):
    """
    The output of "getenforce" command is in one of "Enforcing", "Permissive",
    or "Disabled", so we can return the content directly.
    """
    return {"status": context.content[0]}



from insights.parsers.grub_conf import Grub1Config, Grub2Config
from insights.tests import context_wrap

BAD_DEFAULT_1 = """
#boot=/dev/sda

default=${last}
timeout=0
splashimage=(hd0,0)/grub/splash.xpm.gz
hiddenmenu
title Red Hat Enterprise Linux Server (2.6.32-431.11.2.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.11.2.el6.x86_64 crashkernel=128M@8M rhgb quiet
""".strip()

GOOD_OFFSET_4 = """
#boot=/dev/sda

default=1
timeout=0
splashimage=(hd0,0)/grub/splash.xpm.gz
hiddenmenu
title Red Hat Enterprise Linux Server (2.6.32-431.17.1.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.17.1.el6.x86_64 crashkernel=  rhgb quiet
title Red Hat Enterprise Linux Server (2.6.32-431.11.2.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.11.2.el6.x86_64 crashkernel=128M@32M rhgb quiet
""".strip()

GOOD_OFFSET_3 = """
#boot=/dev/sda
default=1
timeout=0
splashimage=(hd0,0)/grub/splash.xpm.gz
hiddenmenu
title Red Hat Enterprise Linux Server (2.6.32-431.17.1.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.17.1.el6.x86_64 crashkernel=128M@0  rhgb quiet
title Red Hat Enterprise Linux Server (2.6.32-431.11.2.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.11.2.el6.x86_64 crashkernel=128M@0 rhgb quiet
""".strip()

GOOD_OFFSET_2 = """
#boot=/dev/sda

default=1
timeout=0
splashimage=(hd0,0)/grub/splash.xpm.gz
hiddenmenu
title Red Hat Enterprise Linux Server (2.6.32-431.17.1.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.17.1.el6.x86_64 crashkernel=128M@0M  rhgb quiet
title Red Hat Enterprise Linux Server (2.6.32-431.11.2.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.11.2.el6.x86_64 crashkernel=128M@0M rhgb quiet
""".strip()

GOOD_OFFSET_1 = """
#boot=/dev/sda
default=0
timeout=0
splashimage=(hd0,0)/grub/splash.xpm.gz
hiddenmenu
title Red Hat Enterprise Linux Server (2.6.32-431.17.1.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.17.1.el6.x86_64 crashkernel=128M rhgb quiet
title Red Hat Enterprise Linux Server (2.6.32-431.11.2.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.11.2.el6.x86_64 crashkernel=128M rhgb quiet
""".strip()

BAD_OFFSET = """
#boot=/dev/sda

default=1
timeout=0
splashimage=(hd0,0)/grub/splash.xpm.gz
hiddenmenu
title Red Hat Enterprise Linux Server (2.6.32-431.17.1.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.17.1.el6.x86_64 crashkernel=  rhgb quiet
title Red Hat Enterprise Linux Server (2.6.32-431.11.2.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.11.2.el6.x86_64 crashkernel=128M@16M rhgb quiet
""".strip()

NOMATCH_MEMORY = """
#boot=/dev/sda
default=1
timeout=0
splashimage=(hd0,0)/grub/splash.xpm.gz
hiddenmenu
title Red Hat Enterprise Linux Server (2.6.32-431.17.1.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.17.1.el6.x86_64 rhgb quiet
title Red Hat Enterprise Linux Server (2.6.32-431.11.2.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.11.2.el6.x86_64 crashkernel=128M@M rhgb quiet
""".strip()

NOMATCH_CRASH_PARAM = """
#boot=/dev/sda
default=1
timeout=0
splashimage=(hd0,0)/grub/splash.xpm.gz
hiddenmenu
title Red Hat Enterprise Linux Server (2.6.32-431.17.1.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.17.1.el6.x86_64 rhgb quiet
title Red Hat Enterprise Linux Server (2.6.32-431.11.2.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.11.2.el6.x86_64 rhgb quiet
""".strip()

IOMMU_OFF = """
title Red Hat Enterprise Linux Server (2.6.32-431.17.1.el6.x86_64)
    kernel /vmlinuz-2.6.32-279.el6.x86_64 ro root=/dev/mapper/vg00-lv00 intel_iommu=off rd_LVM_LV=vg00/lv00 crashkernel=256M@16M
"""
IOMMU_MISSING = """
title Red Hat Enterprise Linux Server (2.6.32-431.17.1.el6.x86_64)
    kernel /vmlinuz-2.6.32-279.el6.x86_64 ro root=/dev/mapper/vg00-lv00 rd_LVM_LV=vg00/lv00 crashkernel=256M@16M
"""
IOMMU_ON = """
title Red Hat Enterprise Linux Server (2.6.32-431.17.1.el6.x86_64)
    kernel /vmlinuz-2.6.32-279.el6.x86_64 ro root=/dev/mapper/vg00-lv00 intel_iommu=on rd_LVM_LV=vg00/lv00 crashkernel=256M@16M
"""

IOMMU2_OFF = """
menuentry 'Red Hat Enterprise Linux Workstation (3.10.0-327.36.3.el7.x86_64) 7.2 (Maipo)' --class red --class gnu-linux --class gnu --class os --unrestricted $menuentry_id_option 'gnulinux-3.10.0-123.13.2.el7.x86_64-advanced-fbff9f50-62c3-484e-bca5-d53f672cda7c' {
    linux16 /vmlinuz-3.10.0-327.36.3.el7.x86_64 root=/dev/RHEL7CSB/Root ro rd.lvm.lv=RHEL7CSB/Root rd.luks.uuid=luks-96c66446-77fd-4431-9508-f6912bd84194 intel_iommu=off crashkernel=128M@16M rd.lvm.lv=RHEL7CSB/Swap vconsole.font=latarcyrheb-sun16 rhgb quiet LANG=en_GB.utf8
    initrd16 /initramfs-3.10.0-327.36.3.el7.x86_64.img
}
"""
IOMMU2_MISSING = """
menuentry 'Red Hat Enterprise Linux Workstation (3.10.0-327.36.3.el7.x86_64) 7.2 (Maipo)' --class red --class gnu-linux --class gnu --class os --unrestricted $menuentry_id_option 'gnulinux-3.10.0-123.13.2.el7.x86_64-advanced-fbff9f50-62c3-484e-bca5-d53f672cda7c' {
    linux16 /vmlinuz-3.10.0-327.36.3.el7.x86_64 root=/dev/RHEL7CSB/Root ro rd.lvm.lv=RHEL7CSB/Root rd.luks.uuid=luks-96c66446-77fd-4431-9508-f6912bd84194 crashkernel=128M@16M rd.lvm.lv=RHEL7CSB/Swap vconsole.font=latarcyrheb-sun16 rhgb quiet LANG=en_GB.utf8
    initrd16 /initramfs-3.10.0-327.36.3.el7.x86_64.img
}
"""
IOMMU2_ON = """
menuentry 'Red Hat Enterprise Linux Workstation (3.10.0-327.36.3.el7.x86_64) 7.2 (Maipo)' --class red --class gnu-linux --class gnu --class os --unrestricted $menuentry_id_option 'gnulinux-3.10.0-123.13.2.el7.x86_64-advanced-fbff9f50-62c3-484e-bca5-d53f672cda7c' {
    linux16 /vmlinuz-3.10.0-327.36.3.el7.x86_64 root=/dev/RHEL7CSB/Root ro rd.lvm.lv=RHEL7CSB/Root rd.luks.uuid=luks-96c66446-77fd-4431-9508-f6912bd84194 intel_iommu=on crashkernel=128M@16M rd.lvm.lv=RHEL7CSB/Swap vconsole.font=latarcyrheb-sun16 rhgb quiet LANG=en_GB.utf8
    initrd16 /initramfs-3.10.0-327.36.3.el7.x86_64.img
}
"""

GRUB2_CONFIG = """
### BEGIN /etc/grub.d/00_header ###
set pager=1
/
if [ -s $prefix/grubenv ]; then
  load_env
fi
#[...]
if [ x"${feature_menuentry_id}" = xy ]; then
  menuentry_id_option="--id"
else
  menuentry_id_option=""
fi
#[...]
### BEGIN /etc/grub.d/10_linux ###
menuentry 'Red Hat Enterprise Linux Workstation (3.10.0-327.36.3.el7.x86_64) 7.2 (Maipo)' --class red --class gnu-linux --class gnu --class os --unrestricted $menuentry_id_option 'gnulinux-3.10.0-123.13.2.el7.x86_64-advanced-fbff9f50-62c3-484e-bca5-d53f672cda7c' {
    load_video
    set gfxpayload=keep
    insmod gzio
    insmod part_msdos
    insmod ext2
    set root='hd0,msdos1'
    if [ x$feature_platform_search_hint = xy ]; then
      search --no-floppy --fs-uuid --set=root --hint-bios=hd0,msdos1 --hint-efi=hd0,msdos1 --hint-baremetal=ahci0,msdos1 --hint='hd0,msdos1'  1184ab74-77b5-4cfa-81d3-fb87b0457577
    else
      search --no-floppy --fs-uuid --set=root 1184ab74-77b5-4cfa-81d3-fb87b0457577
    fi
    linux16 /vmlinuz-3.10.0-327.36.3.el7.x86_64 root=/dev/RHEL7CSB/Root ro rd.lvm.lv=RHEL7CSB/Root rd.luks.uuid=luks-96c66446-77fd-4431-9508-f6912bd84194 crashkernel=128M@16M rd.lvm.lv=RHEL7CSB/Swap vconsole.font=latarcyrheb-sun16 rhgb quiet LANG=en_GB.utf8
    initrd16 /initramfs-3.10.0-327.36.3.el7.x86_64.img
}
menuentry 'Red Hat Enterprise Linux Workstation (3.10.0-267.el7.x86_64) 7.1 (Maipo)' --class red --class gnu-linux --class gnu --class os --unrestricted $menuentry_id_option 'gnulinux-3.10.0-123.13.2.el7.x86_64-advanced-fbff9f50-62c3-484e-bca5-d53f672cda7c' {
    load_video
    set gfxpayload=keep
    insmod gzio
    insmod part_msdos
    insmod ext2
    set root='hd0,msdos1'
    if [ x$feature_platform_search_hint = xy ]; then
      search --no-floppy --fs-uuid --set=root --hint-bios=hd0,msdos1 --hint-efi=hd0,msdos1 --hint-baremetal=ahci0,msdos1 --hint='hd0,msdos1'  1184ab74-77b5-4cfa-81d3-fb87b0457577
    else
      search --no-floppy --fs-uuid --set=root 1184ab74-77b5-4cfa-81d3-fb87b0457577
    fi
    linux16 /vmlinuz-3.10.0-267.el7.x86_64 root=/dev/RHEL7CSB/Root ro rd.lvm.lv=RHEL7CSB/Root rd.luks.uuid=luks-96c66446-77fd-4431-9508-f6912bd84194 crashkernel=auto  vconsole.keymap=us rd.lvm.lv=RHEL7CSB/Swap vconsole.font=latarcyrheb-sun16 rhgb quiet LANG=en_GB.utf8
    initrd16 /initramfs-3.10.0-267.el7.x86_64.img
}
menuentry 'Red Hat Enterprise Linux Workstation (3.10.0-230.el7synaptics.1186112.1186106.2.x86_64) 7.1 (Maipo)' --class red --class gnu-linux --class gnu --class os --unrestricted $menuentry_id_option 'gnulinux-3.10.0-123.13.2.el7.x86_64-advanced-fbff9f50-62c3-484e-bca5-d53f672cda7c' {
    load_video
    set gfxpayload=keep
    insmod gzio
    insmod part_msdos
    insmod ext2
    set root='hd0,msdos1'
    if [ x$feature_platform_search_hint = xy ]; then
      search --no-floppy --fs-uuid --set=root --hint-bios=hd0,msdos1 --hint-efi=hd0,msdos1 --hint-baremetal=ahci0,msdos1 --hint='hd0,msdos1'  1184ab74-77b5-4cfa-81d3-fb87b0457577
    else
      search --no-floppy --fs-uuid --set=root 1184ab74-77b5-4cfa-81d3-fb87b0457577
    fi
    linux16 /vmlinuz-3.10.0-230.el7synaptics.1186112.1186106.2.x86_64 root=/dev/RHEL7CSB/Root ro rd.lvm.lv=RHEL7CSB/Root rd.luks.uuid=luks-96c66446-77fd-4431-9508-f6912bd84194 crashkernel=auto  vconsole.keymap=us rd.lvm.lv=RHEL7CSB/Swap vconsole.font=latarcyrheb-sun16 rhgb quiet LANG=en_US.UTF-8
    initrd16 /initramfs-3.10.0-230.el7synaptics.1186112.1186106.2.x86_64.img
}
menuentry 'Red Hat Enterprise Linux Workstation, with Linux 0-rescue-71483baa33934d94a7804a398fed6241' --class red --class gnu-linux --class gnu --class os --unrestricted $menuentry_id_option 'gnulinux-0-rescue-71483baa33934d94a7804a398fed6241-advanced-fbff9f50-62c3-484e-bca5-d53f672cda7c' {
    load_video
    insmod gzio
    insmod part_msdos
    insmod ext2
    set root='hd0,msdos1'
    if [ x$feature_platform_search_hint = xy ]; then
      search --no-floppy --fs-uuid --set=root --hint-bios=hd0,msdos1 --hint-efi=hd0,msdos1 --hint-baremetal=ahci0,msdos1 --hint='hd0,msdos1'  1184ab74-77b5-4cfa-81d3-fb87b0457577
    else
      search --no-floppy --fs-uuid --set=root 1184ab74-77b5-4cfa-81d3-fb87b0457577
    fi
    linux16 /vmlinuz-0-rescue-71483baa33934d94a7804a398fed6241 root=UUID=fbff9f50-62c3-484e-bca5-d53f672cda7c ro rd.lvm.lv=RHEL7CSB/Root rd.luks.uuid=luks-96c66446-77fd-4431-9508-f6912bd84194 crashkernel=auto  vconsole.keymap=us rd.lvm.lv=RHEL7CSB/Swap vconsole.font=latarcyrheb-sun16 rhgb quiet
    initrd16 /initramfs-0-rescue-71483baa33934d94a7804a398fed6241.img
}

"""


MODULE_TEST = """
### BEGIN /etc/grub.d/00_header ###
set pager=1
/
if [ -s $prefix/grubenv ]; then
  load_env
fi
#[...]
if [ x"${feature_menuentry_id}" = xy ]; then
  menuentry_id_option="--id"
else
  menuentry_id_option=""
fi
#[...]
### BEGIN /etc/grub.d/10_linux ###
menuentry 'Red Hat Enterprise Linux Workstation (3.10.0-327.36.3.el7.x86_64) 7.2 (Maipo)' --class red --class gnu-linux --class gnu --class os --unrestricted $menuentry_id_option 'gnulinux-3.10.0-123.13.2.el7.x86_64-advanced-fbff9f50-62c3-484e-bca5-d53f672cda7c' {
    load_video
    set gfxpayload=keep
    insmod gzio
    insmod part_msdos
    insmod ext2
    set root='hd0,msdos1'
    if [ x$feature_platform_search_hint = xy ]; then
      search --no-floppy --fs-uuid --set=root --hint-bios=hd0,msdos1 --hint-efi=hd0,msdosyy1 --hint-baremetal=ahci0,msdos1 --hint='hd0,msdos1'  1184ab74-77b5-4cfa-81d3-fb87b0457577
    else
      search --no-floppy --fs-uuid --set=root 1184ab74-77b5-4cfa-81d3-fb87b0457577
    fi
    linux16 /vmlinuz-3.10.0-327.36.3.el7.x86_64 root=/dev/RHEL7CSB/Root ro rd.lvm.lv=RHEL7CSB/Root rd.luks.uuid=luks-96c66446-77fd-4431-9508-f6912bd84194 crashkernel=128M@16M rd.lvm.lv=RHEL7CSB/Swap vconsole.font=latarcyrheb-sun16 rhgb quiet LANG=en_GB.utf8
    initrd16 /initramfs-3.10.0-327.36.3.el7.x86_64.img
"""


def test_mod_internal():
    config = Grub2Config(context_wrap(MODULE_TEST))
    assert config
    assert type(config.kernel_initrds) == dict
    # Why is this two separate lists and not a list of dicts?
    assert 'grub_initrds' in config.kernel_initrds
    assert config.kernel_initrds['grub_initrds'] == ['initramfs-3.10.0-327.36.3.el7.x86_64.img']
    assert 'grub_kernels' in config.kernel_initrds
    assert config.kernel_initrds['grub_kernels'] == ['vmlinuz-3.10.0-327.36.3.el7.x86_64']


def test_kdump_iommu_enabled():
    assert Grub1Config(context_wrap(IOMMU_OFF)).is_kdump_iommu_enabled is False
    assert Grub1Config(context_wrap(IOMMU_MISSING)).is_kdump_iommu_enabled is False
    assert Grub1Config(context_wrap(IOMMU_ON)).is_kdump_iommu_enabled is True
    assert Grub2Config(context_wrap(IOMMU2_OFF)).is_kdump_iommu_enabled is False
    assert Grub2Config(context_wrap(IOMMU2_MISSING)).is_kdump_iommu_enabled is False
    assert Grub2Config(context_wrap(IOMMU2_ON)).is_kdump_iommu_enabled is True


"""
#boot=/dev/sda
default 0
timeout=0
splashimage=(hd0,0)/grub/splash.xpm.gz
hiddenmenu
title Red Hat Enterprise Linux Server (2.6.32-431.17.1.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.17.1.el6.x86_64 crashkernel=128M rhgb quiet
title Red Hat Enterprise Linux Server (2.6.32-431.11.2.el6.x86_64)
        kernel /vmlinuz-2.6.32-431.11.2.el6.x86_64 crashkernel=128M rhgb quiet
"""


def test_grub1_config():
    config = Grub1Config(context_wrap(GOOD_OFFSET_1))
    assert config
    assert 'configs' in config
    assert 'title' in config
    assert 'menuentry' not in config

    assert len(config['configs']) == 4
    assert config['configs']['default'] == ['0']
    assert config['configs']['timeout'] == ['0']
    assert config['configs']['splashimage'] == ['(hd0,0)/grub/splash.xpm.gz']
    assert config['configs']['hiddenmenu'] == ['']

    assert len(config['title']) == 2
    assert len(config['title'][0]) == 2
    assert config['title'][0]['title'] == 'Red Hat Enterprise Linux Server (2.6.32-431.17.1.el6.x86_64)'
    assert config['title'][0]['kernel'][0] == '/vmlinuz-2.6.32-431.17.1.el6.x86_64 crashkernel=128M rhgb quiet'
    assert len(config['title'][1]) == 2
    assert config['title'][1]['title'] == 'Red Hat Enterprise Linux Server (2.6.32-431.11.2.el6.x86_64)'
    assert config['title'][1]['kernel'][-1] == '/vmlinuz-2.6.32-431.11.2.el6.x86_64 crashkernel=128M rhgb quiet'

    assert config.is_kdump_iommu_enabled is False
    assert type(config.kernel_initrds) == dict
    # Why is this two separate lists and not a list of dicts?
    # Because this config has been deliberately cut down and is not a
    # real GRUB configuration, it can't find the initrds.
    assert 'grub_initrds' in config.kernel_initrds
    assert config.kernel_initrds['grub_initrds'] == []
    assert 'grub_kernels' in config.kernel_initrds
    assert config.kernel_initrds['grub_kernels'] == [
        'vmlinuz-2.6.32-431.17.1.el6.x86_64',
        'vmlinuz-2.6.32-431.11.2.el6.x86_64'
    ]


def test_grub2_config():
    config = Grub2Config(context_wrap(GRUB2_CONFIG))
    assert config
    assert 'configs' in config
    assert 'title' not in config
    assert 'menuentry' in config

    # Current parsing code is near useless for pulling information out of
    # GRUB2 config, as the file is actually a bash script.  So here we
    # don't do any testing of the junk in 'configs'.
    assert len(config['configs']) > 0

    assert len(config['menuentry']) == 4
    """
    menuentry 'Red Hat Enterprise Linux Workstation (3.10.0-327.36.3.el7.x86_64) 7.2 (Maipo)' --class red --class gnu-linux --class gnu --class os --unrestricted $menuentry_id_option 'gnulinux-3.10.0-123.13.2.el7.x86_64-advanced-fbff9f50-62c3-484e-bca5-d53f672cda7c' {
        load_video
        set gfxpayload=keep
        insmod gzio
        insmod part_msdos
        insmod ext2
        set root='hd0,msdos1'
        if [ x$feature_platform_search_hint = xy ]; then
          search --no-floppy --fs-uuid --set=root --hint-bios=hd0,msdos1 --hint-efi=hd0,msdos1 --hint-baremetal=ahci0,msdos1 --hint='hd0,msdos1'  1184ab74-77b5-4cfa-81d3-fb87b0457577
        else
          search --no-floppy --fs-uuid --set=root 1184ab74-77b5-4cfa-81d3-fb87b0457577
        fi
        linux16 /vmlinuz-3.10.0-327.36.3.el7.x86_64 root=/dev/RHEL7CSB/Root ro rd.lvm.lv=RHEL7CSB/Root rd.luks.uuid=luks-96c66446-77fd-4431-9508-f6912bd84194 crashkernel=auto  vconsole.keymap=us rd.lvm.lv=RHEL7CSB/Swap vconsole.font=latarcyrheb-sun16 rhgb quiet LANG=en_GB.utf8
        initrd16 /initramfs-3.10.0-327.36.3.el7.x86_64.img
    }
    """
    # The current parsing code does a very bad job of reading menu
    # entry configuration.  Test a few things but we hope for better.
    assert len(config['menuentry'][0]) == 6
    assert config['menuentry'][0]['menuentry'] == "'Red Hat Enterprise Linux Workstation (3.10.0-327.36.3.el7.x86_64) 7.2 (Maipo)' --class red --class gnu-linux --class gnu --class os --unrestricted $menuentry_id_option 'gnulinux-3.10.0-123.13.2.el7.x86_64-advanced-fbff9f50-62c3-484e-bca5-d53f672cda7c'"
    assert config['menuentry'][0]['load_video'] == ['']
    assert config['menuentry'][0]['initrd16'] == ['/initramfs-3.10.0-327.36.3.el7.x86_64.img']

    assert type(config.kernel_initrds) == dict
    assert 'grub_initrds' in config.kernel_initrds
    assert config.kernel_initrds['grub_initrds'] == [
        'initramfs-3.10.0-327.36.3.el7.x86_64.img',
        'initramfs-3.10.0-267.el7.x86_64.img',
        'initramfs-3.10.0-230.el7synaptics.1186112.1186106.2.x86_64.img',
        'initramfs-0-rescue-71483baa33934d94a7804a398fed6241.img'
    ]
    assert 'grub_kernels' in config.kernel_initrds
    assert config.kernel_initrds['grub_kernels'] == [
        'vmlinuz-3.10.0-327.36.3.el7.x86_64',
        'vmlinuz-3.10.0-267.el7.x86_64',
        'vmlinuz-3.10.0-230.el7synaptics.1186112.1186106.2.x86_64',
        'vmlinuz-0-rescue-71483baa33934d94a7804a398fed6241'
    ]



from insights.parsers.networkmanager_config import NetworkManagerConfig
from insights.parsers import networkmanager_config
from insights.tests import context_wrap
import doctest

NETWORKMANAGER_CONF = """
# Configuration file for NetworkManager.
#
# See "man 5 NetworkManager.conf" for details.
#
# The directories /usr/lib/NetworkManager/conf.d/ and /var/run/NetworkManager/conf.d/
# can contain additional configuration snippets installed by packages. These files are
# read before NetworkManager.conf and have thus lowest priority.
# The directory /etc/NetworkManager/conf.d/ can contain additional configuration
# snippets. Those snippets are merged last and overwrite the settings from this main
# file.
#
# The files within one conf.d/ directory are read in asciibetical order.
#
# If /etc/NetworkManager/conf.d/ contains a file with the same name as
# /usr/lib/NetworkManager/conf.d/, the latter file is shadowed and thus ignored.
# Hence, to disable loading a file from /usr/lib/NetworkManager/conf.d/ you can
# put an empty file to /etc with the same name. The same applies with respect
# to the directory /var/run/NetworkManager/conf.d where files in /var/run shadow
# /usr/lib and are themselves shadowed by files under /etc.
#
# If two files define the same key, the one that is read afterwards will overwrite
# the previous one.

[main]
#plugins=ifcfg-rh,ibft
dhcp=dhclient


[logging]
# When debugging NetworkManager, enabling debug logging is of great help.
#
# Logfiles contain no passwords and little sensitive information. But please
# check before posting the file online. You can also personally hand over the
# logfile to a NM developer to treat it confidential. Meet us on #nm on freenode.
# Please post full logfiles except minimal modifications of private data.
#
# You can also change the log-level at runtime via
#   $ nmcli general logging level TRACE domains ALL
# However, usually it's cleaner to enable debug logging
# in the configuration and restart NetworkManager so that
# debug logging is enabled from the start.
#
# You will find the logfiles in syslog, for example via
#   $ journalctl -u NetworkManager
#
# Note that debug logging of NetworkManager can be quite verbose. Some messages
# might be rate-limited by the logging daemon (see RateLimitIntervalSec, RateLimitBurst
# in man journald.conf).
#
#level=TRACE
#domains=ALL
"""

NETWORKMANAGER_CONF_NOTMATCH = """
# Configuration file for NetworkManager.
#
# See "man 5 NetworkManager.conf" for details.
#
# The directories /usr/lib/NetworkManager/conf.d/ and /var/run/NetworkManager/conf.d/
# can contain additional configuration snippets installed by packages. These files are
# read before NetworkManager.conf and have thus lowest priority.
# The directory /etc/NetworkManager/conf.d/ can contain additional configuration
# snippets. Those snippets are merged last and overwrite the settings from this main
# file.
#
# The files within one conf.d/ directory are read in asciibetical order.
#
# If /etc/NetworkManager/conf.d/ contains a file with the same name as
# /usr/lib/NetworkManager/conf.d/, the latter file is shadowed and thus ignored.
# Hence, to disable loading a file from /usr/lib/NetworkManager/conf.d/ you can
# put an empty file to /etc with the same name. The same applies with respect
# to the directory /var/run/NetworkManager/conf.d where files in /var/run shadow
# /usr/lib and are themselves shadowed by files under /etc.
#
# If two files define the same key, the one that is read afterwards will overwrite
# the previous one.

[logging]
# When debugging NetworkManager, enabling debug logging is of great help.
#
# Logfiles contain no passwords and little sensitive information. But please
# check before posting the file online. You can also personally hand over the
# logfile to a NM developer to treat it confidential. Meet us on #nm on freenode.
# Please post full logfiles except minimal modifications of private data.
#
# You can also change the log-level at runtime via
#   $ nmcli general logging level TRACE domains ALL
# However, usually it's cleaner to enable debug logging
# in the configuration and restart NetworkManager so that
# debug logging is enabled from the start.
#
# You will find the logfiles in syslog, for example via
#   $ journalctl -u NetworkManager
#
# Note that debug logging of NetworkManager can be quite verbose. Some messages
# might be rate-limited by the logging daemon (see RateLimitIntervalSec, RateLimitBurst
# in man journald.conf).
#
#level=TRACE
domains=ALL
"""


def test_networkmanager_config_match():
    result = NetworkManagerConfig(context_wrap(NETWORKMANAGER_CONF))
    assert result.get('main', 'dhcp') == 'dhclient'


def test_networkmanager_config_notmatch():
    result = NetworkManagerConfig(context_wrap(NETWORKMANAGER_CONF_NOTMATCH))
    assert result.has_option('main', 'dhcp') is False


def test_networkmanager_config_doc_examples():
    env = {
        'networkmanager_config_obj': NetworkManagerConfig(context_wrap(NETWORKMANAGER_CONF)),
    }
    failed, total = doctest.testmod(networkmanager_config, globs=env)
    assert failed == 0



"""
IpsecConf parser - file ``/etc/ipsec.conf``
===========================================

IpsecConf parser the file /etc/ipsec.conf about
the configuration and control information
for the Libreswan IPsec subsystem.
"""

from collections import defaultdict

from insights.specs import Specs
from insights.core import CommandParser
from .. import parser, get_active_lines
from insights.parsers import SkipException


@parser(Specs.ipsec_conf)
class IpsecConf(CommandParser, dict):
    """
    Class for parsing the file ``/etc/ipsec.conf`` about the configuration
    and control information for the Libreswan IPsec subsystem

    Raises:
        SkipException: When content is empty or cannot be parsed.

    Sample output of this command is::

        # /etc/ipsec.conf - Libreswan IPsec configuration file
        #
        # see 'man ipsec.conf' and 'man pluto' for more information
        #
        # For example configurations and documentation, see https://libreswan.org/wiki/

        config setup
                # plutodebug="control parsing"
                # plutodebug="all crypt"
                plutodebug=none
                # It seems that T-Mobile in the US and Rogers/Fido in Canada are
                # using 25/8 as "private" address space on their wireless networks.
                # This range has never been announced via BGP (at least up to 2015)
                virtual_private=%v4:10.0.0.0/8,%v4:192.168.0.0/16,%v4:172.16.0.0/12,%v4:25.0.0.0/8,%v4:100.64.0.0/10,%v6:fd00::/8,%v6:fe80::/10

        # if it exists, include system wide crypto-policy defaults
        include /etc/crypto-policies/back-ends/libreswan.config

        # It is best to add your IPsec connections as separate files in /etc/ipsec.d/
        include /etc/ipsec.d/*.conf

    Examples:
        >>> ipsec_conf['config']['setup']['plutodebug'] == 'none'
        True
        >>> ipsec_conf['include']
        ['/etc/crypto-policies/back-ends/libreswan.config', '/etc/ipsec.d/*.conf']
    """

    def parse_content(self, content):
        if not content:
            raise SkipException('No content.')

        ipsec_type, ipsec_name = "", ""
        ipsec_sections = {}

        try:
            for line in get_active_lines(content):
                if line.startswith('include '):
                    include, path = [field.strip() for field in line.split()]
                    array = self.get('include', [])
                    array.append(path)
                    self['include'] = array
                    continue

                if line.startswith(('conn ', 'config ')):
                    ipsec_type, ipsec_name = [field.strip() for field in line.split()]
                    ipsec_sections = self.get(ipsec_type, defaultdict(dict))
                    continue

                if '=' not in line or ipsec_type == "" or ipsec_name == "":
                    # skip the options that don't within a section
                    continue

                key, value = [field.strip() for field in line.split('=')]
                ipsec_sections[ipsec_name][key] = value
                self[ipsec_type] = ipsec_sections

        except ValueError:
            raise SkipException('Syntax error')



# coding: utf-8

from __future__ import print_function, absolute_import, division, unicode_literals

from ...ruamel.yaml.compat import text_type
from ...ruamel.yaml.anchor import Anchor

if False:  # MYPY
    from typing import Text, Any, Dict, List  # NOQA

__all__ = [
    'ScalarString',
    'LiteralScalarString',
    'FoldedScalarString',
    'SingleQuotedScalarString',
    'DoubleQuotedScalarString',
    'PlainScalarString',
    # PreservedScalarString is the old name, as it was the first to be preserved on rt,
    # use LiteralScalarString instead
    'PreservedScalarString',
]


class ScalarString(text_type):
    __slots__ = Anchor.attrib

    def __new__(cls, *args, **kw):
        # type: (Any, Any) -> Any
        anchor = kw.pop('anchor', None)  # type: ignore
        ret_val = text_type.__new__(cls, *args, **kw)  # type: ignore
        if anchor is not None:
            ret_val.yaml_set_anchor(anchor, always_dump=True)
        return ret_val

    def replace(self, old, new, maxreplace=-1):
        # type: (Any, Any, int) -> Any
        return type(self)((text_type.replace(self, old, new, maxreplace)))

    @property
    def anchor(self):
        # type: () -> Any
        if not hasattr(self, Anchor.attrib):
            setattr(self, Anchor.attrib, Anchor())
        return getattr(self, Anchor.attrib)

    def yaml_anchor(self, any=False):
        # type: (bool) -> Any
        if not hasattr(self, Anchor.attrib):
            return None
        if any or self.anchor.always_dump:
            return self.anchor
        return None

    def yaml_set_anchor(self, value, always_dump=False):
        # type: (Any, bool) -> None
        self.anchor.value = value
        self.anchor.always_dump = always_dump


class LiteralScalarString(ScalarString):
    __slots__ = 'comment'  # the comment after the | on the first line

    style = '|'

    def __new__(cls, value, anchor=None):
        # type: (Text, Any) -> Any
        return ScalarString.__new__(cls, value, anchor=anchor)


PreservedScalarString = LiteralScalarString


class FoldedScalarString(ScalarString):
    __slots__ = ('fold_pos', 'comment')  # the comment after the > on the first line

    style = '>'

    def __new__(cls, value, anchor=None):
        # type: (Text, Any) -> Any
        return ScalarString.__new__(cls, value, anchor=anchor)


class SingleQuotedScalarString(ScalarString):
    __slots__ = ()

    style = "'"

    def __new__(cls, value, anchor=None):
        # type: (Text, Any) -> Any
        return ScalarString.__new__(cls, value, anchor=anchor)


class DoubleQuotedScalarString(ScalarString):
    __slots__ = ()

    style = '"'

    def __new__(cls, value, anchor=None):
        # type: (Text, Any) -> Any
        return ScalarString.__new__(cls, value, anchor=anchor)


class PlainScalarString(ScalarString):
    __slots__ = ()

    style = ''

    def __new__(cls, value, anchor=None):
        # type: (Text, Any) -> Any
        return ScalarString.__new__(cls, value, anchor=anchor)


def preserve_literal(s):
    # type: (Text) -> Text
    return LiteralScalarString(s.replace('\r\n', '\n').replace('\r', '\n'))


def walk_tree(base, map=None):
    # type: (Any, Any) -> None
    """
    the routine here walks over a simple yaml tree (recursing in
    dict values and list items) and converts strings that
    have multiple lines to literal scalars

    You can also provide an explicit (ordered) mapping for multiple transforms
    (first of which is executed):
        map = ruamel.yaml.compat.ordereddict
        map['\n'] = preserve_literal
        map[':'] = SingleQuotedScalarString
        walk_tree(data, map=map)
    """
    from ...ruamel.yaml.compat import string_types
    from ...ruamel.yaml.compat import MutableMapping, MutableSequence  # type: ignore

    if map is None:
        map = {'\n': preserve_literal}

    if isinstance(base, MutableMapping):
        for k in base:
            v = base[k]  # type: Text
            if isinstance(v, string_types):
                for ch in map:
                    if ch in v:
                        base[k] = map[ch](v)
                        break
            else:
                walk_tree(v, map=map)
    elif isinstance(base, MutableSequence):
        for idx, elem in enumerate(base):
            if isinstance(elem, string_types):
                for ch in map:
                    if ch in elem:  # type: ignore
                        base[idx] = map[ch](elem)
                        break
            else:
                walk_tree(elem, map=map)



"""
Lsof - command ``/usr/sbin/lsof``
=================================

This parser reads the output of the ``/usr/sbin/lsof`` command and makes each
line available as a dictionary keyed on the fields in the lsof output (with
names in upper case).

Because of the large quantity of output from this command, this class is based
on the ``Scannable`` parser class.  There are several ways to use this:

* If you simply want to know whether a search matched, use the ``any`` method.
* If you want all lines that match, use the ``collect`` method.

The way these scanner functions work is:

1. You provide a function, which returns True if a match is found, and an
   attribute name.
2. The parser runs every scanner function across every line of data read and
   successfully parsed.
3. The attribute then contains the result of the match (True or False in the
   case of the ``any`` method, or the list of matching rows in the case of
   the ``collect`` method.

As an easier way of finding all the lines that match by key=value pairs, use
the ``collect_keys`` method, giving the name of the scanner attribute to set
and one or more 'key=value' pairs in the method call.  This then returns the
list of rows for which the data in all those keywords matched the respective
given values.  (**Note**: the ``SIZE/OFF`` column is searched for using the
key ``SIZE_OFF`` - see example below)

Sample output::

    COMMAND     PID  TID           USER   FD      TYPE             DEVICE  SIZE/OFF       NODE NAME
    systemd       1                root  cwd       DIR              253,1      4096        128 /
    systemd       1                root  rtd       DIR              253,1      4096        128 /
    systemd       1                root  txt       REG              253,1   1230920    1440410 /usr/lib/systemd/systemd
    systemd       1                root  mem       REG              253,1     37152  135529970 /usr/lib64/libnss_sss.so.2
    abrt-watc  8619                root    0r      CHR                1,3       0t0       4674 /dev/null
    wpa_suppl   641                root    0r      CHR                1,3       0t0       4674 /dev/null
    polkitd     642             polkitd    0u      CHR                1,3       0t0       4674 /dev/null
    polkitd     642             polkitd    1u      CHR                1,3       0t0       4674 /dev/null

Examples:

    >>> Lsof.any('systemd_commands', lambda x: 'systemd' in x['COMMAND'])
    >>> Lsof.collect('polkitd_user', lambda x: x['USER'] == 'polkitd')
    >>> Lsof.collect_keys('root_stdin', USER='root', FD='0r', SIZE_OFF='0t0')
    >>> l = shared[Lsof]
    >>> l.systemd_commands
    True
    >>> len(l.polkitd_user)
    2
    >>> l.polkitd_user[0]['DEVICE']
    '1,3'
    >>> len(l.root_stdin)
    2
    >>> l.root_stdin[0]['COMMAND']
    'abrt-watc'

"""

from insights.core.dr import SkipComponent
from .. import add_filter, Scannable, parser, CommandParser
from insights.specs import Specs

add_filter(Specs.lsof, ['COMMAND'])


@parser(Specs.lsof)
class Lsof(CommandParser, Scannable):
    """
    A parser for the output of ``/usr/sbin/lsof`` - determines the column
    widths from the first row and then puts the data in each row into a
    dictionary keyed on the column name and found by the locations of each
    column.  Leading and trailing spaces are stripped from data.
    """

    def _calc_indexes(self, line):
        self.headings = [c.strip() for c in line.split(None)]
        self.indexes = {}
        for heading in self.headings:
            index = line.index(heading)
            # (start, end)
            self.indexes[heading] = (index, index + len(heading))

    def _start(self, content):
        """
        Consumes lines from content until the HEADER is found and processed.
        Returns an iterator over the remaining lines.
        """
        start = len(content)
        for n, line in enumerate(content):
            if 'COMMAND ' in line:
                start = n + 1
                break

        if start >= len(content):
            raise SkipComponent

        content = iter(content[start:])

        self._calc_indexes(line)
        return content

    def _parse_line(self, line):
        """
        Given a line, returns a dictionary for that line. Requires _start to be
        called first.
        """
        # Split NAME seperately as it can have extra whitespaces
        command = line[self.indexes['NAME'][0]:].strip()
        rest = line[:self.indexes['NAME'][0]]
        rdict = dict.fromkeys(self.headings, '')
        rowsplit = [i.strip() for i in rest.split(None, len(self.headings) - 2)]
        if len(rowsplit) < len(self.headings) - 1:
            rowsplit = iter(rowsplit)
            for heading in self.headings[:-1]:
                # Use value if (start, end) index of heading is not empty
                if line[slice(*self.indexes[heading])].strip():
                    rdict[heading] = next(rowsplit, '')
        else:
            rdict = dict(zip(self.headings, rowsplit))
        rdict['NAME'] = command
        return rdict

    def parse(self, content):
        """
        Parse the content for the entire input file.
        """
        for line in self._start(content):
            yield self._parse_line(line)

    @classmethod
    def collect_keys(cls, result_key, **kwargs):
        """
        Store a list of lines having keyword=value matches in the given
        attribute name.

        Keyword argument names that exist as column names in the data are
        searched for in the lines of the file and stored in the ``result_key``
        attribute.  All columns must match (i.e. this is a bolean AND search).

        Call this class method before using the class data.

        Examples:
            collect_keys('root_block_devs', USER='root', TYPE='BLK')
        """
        def scanner(self, obj):
            # Have to set the attribute so it exists, even if it has no rows
            if not hasattr(self, result_key):
                setattr(self, result_key, [])
            # Minor hack - search for 'SIZE/OFF' as 'SIZE_OFF'.
            if 'SIZE_OFF' in kwargs:
                kwargs['SIZE/OFF'] = kwargs['SIZE_OFF']
            # Filter the keywords for the list of column names requested.
            for key in [k for k in kwargs if k in obj]:
                # If we have a keyword whose value doesn't match the given
                # item, then skip out now.
                if obj[key] != kwargs[key]:
                    return
            # OK, save the item now.
            getattr(self, result_key).append(obj)

        cls._scan(result_key, scanner)



"""
parsr is a library for building parsers based on `parsing expression grammars or
PEGs <http://bford.info/pub/lang/peg.pdf>`_.

You build a parser by making subparsers to match simple building blocks like
numbers, strings, symbols, etc. and then composing them to reflect the higher
level structure of your language.

Some means of combination are like those of regular expressions: sequences,
alternatives, repetition, optional matching, etc. However, matching is always
greedy. parsr also allows recursive definitions and the ability to transform
the match of any subparser with a function. The parser can recognize and
interpret its input at the same time.

Here's an example that evaluates arithmetic expressions.

    .. code-block:: python

        from insights.parsr import EOF, Forward, InSet, Many, Number, WS

        def op(args):
            ans, rest = args
            for op, arg in rest:
                if op == "+":
                    ans += arg
                elif op == "-":
                    ans -= arg
                elif op == "*":
                    ans *= arg
                else:
                    ans /= arg
            return ans


        LP = Char("(")
        RP = Char(")")

        expr = Forward()  # Forward declarations allow recursive structure
        factor = WS >> (Number | (LP >> expr << RP)) << WS
        term = (factor + Many(InSet("*/") + factor)).map(op)

        # Notice the funny assignment of Forward definitions.
        expr <= (term + Many(InSet("+-") + term)).map(op)

        evaluate = expr << EOF
"""
from __future__ import print_function
import functools
import logging
import os
import string
import traceback
from bisect import bisect_left
from six import StringIO, with_metaclass

log = logging.getLogger(__name__)


class Node(object):
    """
    Node is the base class of all parsers. It's a generic tree structure with
    each instance containing a list of its children. Its main purpose is to
    simplify pretty printing.
    """
    def __init__(self):
        self.children = []

    def add_child(self, child):
        self.children.append(child)
        return self

    def set_children(self, children):
        self.children = []
        for c in children:
            self.add_child(c)
        return self

    def __repr__(self):
        return self.__class__.__name__


def text_format(tree):
    """
    Converts a PEG into a pretty printed string.
    """
    out = StringIO()
    tab = " " * 2
    seen = set()

    def inner(cur, prefix):
        print(prefix + str(cur), file=out)
        if cur in seen:
            return

        seen.add(cur)

        next_prefix = prefix + tab
        for c in cur.children:
            inner(c, next_prefix)

    inner(tree, "")
    out.seek(0)
    return out.read()


def render(tree):
    """
    Pretty prints a PEG.
    """
    print(text_format(tree))


def _debug_hook(func):
    """
    _debug_hook wraps the process function of every parser. It maintains a
    stack of active parsers during evaluation to help with error reporting and
    prints diagnostic messages for parsers with debug enabled.
    """
    @functools.wraps(func)
    def inner(self, pos, data, ctx):
        if ctx.function_error is not None:
            # no point in continuing...
            raise Exception()
        ctx.parser_stack.append(self)
        if self._debug:
            line = ctx.line(pos) + 1
            col = ctx.col(pos) + 1
            log.debug("Trying {0} at line {1} col {2}".format(self, line, col))
        try:
            res = func(self, pos, data, ctx)
            if self._debug:
                log.debug("Result: {0}".format(res[1]))
            return res
        except:
            if self._debug:
                ps = "-> ".join([str(p) for p in ctx.parser_stack])
                log.debug("Failed: {0}".format(ps))
            raise
        finally:
            ctx.parser_stack.pop()
    return inner


class Backtrack(Exception):
    """
    Mapped or Lifted functions should Backtrack if they want to fail without
    causing parsing to fail.
    """
    def __init__(self, msg):
        super(Backtrack, self).__init__(msg)


class Context(object):
    """
    An instance of Context is threaded through the process call to every
    parser. It stores an indention stack to track hanging indents, a tag stack
    for grammars like xml or apache configuration, the active parser stack for
    error reporting, and accumulated errors for the farthest position reached.
    """
    def __init__(self, lines, src=None):
        self.pos = -1
        self.indents = []
        self.tags = []
        self.src = src
        self.lines = [i for i, x in enumerate(lines) if x == "\n"]
        self.parser_stack = []
        self.errors = []
        self.function_error = None

    def set(self, pos, msg):
        """
        Every parser that encounters an error calls set with the current
        position and a message. If the error is at the farthest position
        reached by any other parser, the active parser stack and message are
        accumulated onto a list of errors for that position. If the position is
        beyond any previous errors, the error list is cleared before the active
        stack and new error are recorded. This is the "farthest failure
        heurstic."
        """
        if pos > self.pos:
            self.errors = []

        if pos >= self.pos:
            self.pos = pos
            self.errors.append((list(self.parser_stack), msg))

    def line(self, pos):
        return bisect_left(self.lines, pos)

    def col(self, pos):
        p = self.line(pos)
        if p == 0:
            return pos
        return (pos - self.lines[p - 1] - 1)


class _ParserMeta(type):
    """
    ParserMeta wraps every parser subclass's process function with the
    ``_debug_hook`` decorator.
    """
    def __init__(cls, name, bases, clsdict):
        orig = getattr(cls, "process")
        setattr(cls, "process", _debug_hook(orig))


class Parser(with_metaclass(_ParserMeta, Node)):
    """
    Parser is the common base class of all Parsers.
    """
    def __init__(self):
        super(Parser, self).__init__()
        self.name = None
        self._debug = False

    def debug(self, d=True):
        """
        Set to ``True`` to enable diagnostic messages before and after the
        parser is invoked.
        """
        self._debug = d
        return self

    @staticmethod
    def _accumulate(first, rest):
        results = [first] if first else []
        if rest:
            results.extend(rest)
        return results

    def sep_by(self, sep):
        """
        Return a parser that matches zero or more instances of the current
        parser separated by instances of the parser sep.
        """
        return Lift(self._accumulate) * Opt(self) * Many(sep >> self)

    def until(self, pred):
        """
        Return an :py:class:`Until` parser that matches zero or more instances
        of the current parser until the pred parser succeeds.
        """
        return Until(self, pred)

    def map(self, func):
        """
        Return a :py:class:`Map` parser that transforms the results of the
        current parser with the function func.
        """
        return Map(self, func)

    def __add__(self, other):
        """
        Return a :py:class:`Sequence` that requires the current parser to be
        followed by the other parser. Additional calls to ``+`` on the returned
        parser will cause it to accumulate "other" onto itself instead of
        creating new Sequences. This has the desirable effect of causing
        sequence matches to be represented as flat lists instead of trees, but
        it can also have unintended consequences if a sequence is used in
        multiple parts of a grammar as the initial element of another sequence.

        :py:class:`Choice`s also accumulate and can lead to similar surprises.
        """
        return Sequence([self, other])

    def __or__(self, other):
        """
        Return a :py:class:`Choice` that requires the current parser or the
        other parser to match. Additional calls to ``|`` on the returned parser
        will cause it to accumulate "other" onto itself instead of creating new
        Choices. This has the desirable effect of making choices more
        efficient, but it can also have unintended consequences if a choice is
        used in multiple parts of a grammar as the initial element of another
        :py:class:`Choice`.
        """
        return Choice([self, other])

    def __lshift__(self, other):
        """
        Return a parser that requires the current parser and the other parser
        but ignores the other parser's result. Both parsers consume input.
        """
        return KeepLeft(self, other)

    def __rshift__(self, other):
        """
        Return a parser that requires the current parser and the other parser
        but ignores the current parser's result. Both parsers consume input.
        """
        return KeepRight(self, other)

    def __and__(self, other):
        """
        Return a parser that requires the current parser and the other parser
        but ignores the current parser's result. Only the current parser
        consumes input.
        """
        return FollowedBy(self, other)

    def __truediv__(self, other):
        """
        Return a parser that requires the current parser but requires the other
        parser to fail. Only the current parser consumes input.
        """
        return NotFollowedBy(self, other)

    def __mod__(self, name):
        """
        Gives the current parser a human friendly name for display in error
        messages and PEG rendering.
        """
        self.name = name
        return self

    def process(self, pos, data, ctx):
        raise NotImplementedError()

    def __call__(self, data, src=None, Ctx=Context):
        """
        Invoke the parser like a function on a regular string of characters.

        Optinally provide an arbitrary external object that will be made available to
        the Context instance. You also can provide a Context subclass if your
        parsers have particular needs not covered by the default
        implementation that provides significant indent and tag stacks.
        """
        data = list(data)
        data.append(None)  # add a terminal so we don't overrun
        ctx = Ctx(data, src=src)

        try:
            _, ret = self.process(0, data, ctx)
            return ret
        except Exception:
            pass

        if ctx.function_error is not None:
            pos, msg = ctx.function_error
            lineno = ctx.line(pos) + 1
            colno = ctx.col(pos) + 1
            msg = "At line {0} column {1}: {2}".format(lineno, colno, msg)
            raise Exception(msg)

        err = StringIO()

        lineno = ctx.line(ctx.pos) + 1
        colno = ctx.col(ctx.pos) + 1
        msg = "At line {0} column {1}:"
        print(msg.format(lineno, colno, ctx.lines), file=err)
        for parsers, msg in ctx.errors:
            names = " -> ".join([p.name for p in parsers if p.name])
            v = data[ctx.pos] or "EOF"
            print(names, file=err)
            print("    {0} Got {1!r}.".format(msg, v), file=err)
        err.seek(0)
        raise Exception(err.read())

    def __repr__(self):
        return self.name or self.__class__.__name__


class AnyChar(Parser):
    def process(self, pos, data, ctx):
        c = data[pos]
        if c is not None:
            return (pos + 1, c)
        msg = "Expected any character."
        ctx.set(pos, msg)
        raise Exception(msg)


class Char(Parser):
    """
    Char matches a single character.

        .. code-block:: python

            a = Char("a")     # parses a single "a"
            val = a("a")      # produces an "a" from the data.
            val = a("b")      # raises an exception

    """
    def __init__(self, char):
        super(Char, self).__init__()
        self.char = char

    def process(self, pos, data, ctx):
        if data[pos] == self.char:
            return (pos + 1, self.char)
        msg = "Expected {0}.".format(self.char)
        ctx.set(pos, msg)
        raise Exception(msg)

    def __repr__(self):
        if self.name is None:
            return "Char({0})".format(self.char)
        return self.name


class InSet(Parser):
    """
    InSet matches any single character from a set.

        .. code-block:: python

            vowel = InSet("aeiou")  # or InSet(set("aeiou"))
            val = vowel("a")  # okay
            val = vowel("e")  # okay
            val = vowel("i")  # okay
            val = vowel("o")  # okay
            val = vowel("u")  # okay
            val = vowel("y")  # raises an exception

    """
    def __init__(self, s, name=None):
        super(InSet, self).__init__()
        self.values = set(s)
        self.name = name

    def process(self, pos, data, ctx):
        c = data[pos]
        if c in self.values:
            return (pos + 1, c)
        msg = "Expected {0}.".format(self)
        ctx.set(pos, msg)
        raise Exception(msg)

    def __repr__(self):
        if self.name is None:
            return "InSet({0!r})".format(sorted(self.values))
        return super(InSet, self).__repr__()


class String(Parser):
    """
    Match one or more characters in a set. Matching is greedy.

        .. code-block:: python

            vowels = String("aeiou")
            val = vowels("a")            # returns "a"
            val = vowels("u")            # returns "u"
            val = vowels("aaeiouuoui")   # returns "aaeiouuoui"
            val = vowels("uoiea")        # returns "uoiea"
            val = vowels("oouieaaea")    # returns "oouieaaea"
            val = vowels("ga")           # raises an exception

    """
    def __init__(self, chars, echars=None, min_length=1):
        super(String, self).__init__()
        self.chars = set(chars)
        self.echars = set(echars) if echars else set()
        self.min_length = min_length

    def process(self, pos, data, ctx):
        results = []
        p = data[pos]
        old = pos
        while p in self.chars or p == "\\":
            if p == "\\" and data[pos + 1] in self.echars:
                results.append(data[pos + 1])
                pos += 2
            elif p in self.chars:
                results.append(p)
                pos += 1
            else:
                break
            p = data[pos]
        if len(results) < self.min_length:
            msg = "Expected {0} of {1}.".format(self.min_length, sorted(self.chars))
            ctx.set(old, msg)
            raise Exception(msg)
        return pos, "".join(results)


class Literal(Parser):
    """
    Match a literal string. The ``value`` keyword lets you return a python
    value instead of the matched input. The ``ignore_case`` keyword makes the
    match case insensitive.

        .. code-block:: python

            lit = Literal("true")
            val = lit("true")  # returns "true"
            val = lit("True")  # raises an exception
            val = lit("one")   # raises an exception

            lit = Literal("true", ignore_case=True)
            val = lit("true")  # returns "true"
            val = lit("TRUE")  # returns "TRUE"
            val = lit("one")   # raises an exception

            t = Literal("true", value=True)
            f = Literal("false", value=False)
            val = t("true")  # returns the boolean True
            val = t("True")  # raises an exception

            val = f("false") # returns the boolean False
            val = f("False") # raises and exception

            t = Literal("true", value=True, ignore_case=True)
            f = Literal("false", value=False, ignore_case=True)
            val = t("true")  # returns the boolean True
            val = t("True")  # returns the boolean True

            val = f("false") # returns the boolean False
            val = f("False") # returns the boolean False
    """
    _NULL = object()

    def __init__(self, chars, value=_NULL, ignore_case=False):
        super(Literal, self).__init__()
        self.chars = chars if not ignore_case else chars.lower()
        self.value = value
        self.ignore_case = ignore_case
        self.name = "Literal{0!r}".format(self.chars)

    def process(self, pos, data, ctx):
        old = pos
        if not self.ignore_case:
            for c in self.chars:
                if data[pos] == c:
                    pos += 1
                else:
                    msg = "Expected {0!r}.".format(self.chars)
                    ctx.set(old, msg)
                    raise Exception(msg)
            return pos, (self.chars if self.value is self._NULL else self.value)
        else:
            result = []
            for c in self.chars:
                if data[pos].lower() == c:
                    result.append(data[pos])
                    pos += 1
                else:
                    msg = "Expected case insensitive {0!r}.".format(self.chars)
                    ctx.set(old, msg)
                    raise Exception(msg)
            return pos, ("".join(result) if self.value is self._NULL else self.value)


class Wrapper(Parser):
    """
    Parser that wraps another parser. This can be used to prevent sequences and
    choices from accidentally accumulating other parsers when used in multiple
    parts of a grammar.
    """
    def __init__(self, parser):
        super(Wrapper, self).__init__()
        self.add_child(parser)

    def process(self, pos, data, ctx):
        return self.children[0].process(pos, data, ctx)


class Mark(object):
    """
    An object created by :py:class:`PosMarker` to capture a value at a position
    in the input. Marks can give more context to a value transformed by mapped
    functions.
    """
    def __init__(self, lineno, col, value):
        self.lineno = lineno
        self.col = col
        self.value = value


class PosMarker(Wrapper):
    """
    Save the line number and column of a subparser by wrapping it in a
    PosMarker. The value of the parser that handled the input as well as the
    initial input position will be returned as a :py:class:`Mark`.
    """
    def process(self, pos, data, ctx):
        lineno = ctx.line(pos) + 1
        col = ctx.col(pos) + 1
        pos, result = super(PosMarker, self).process(pos, data, ctx)
        return pos, Mark(lineno, col, result)


class Sequence(Parser):
    """
    A Sequence requires all of its children to succeed. It returns a list of
    the values they matched.

    Additional uses of ``+`` on the parser will cause it to accumulate parsers
    onto itself instead of creating new Sequences. This has the desirable
    effect of causing sequence results to be represented as flat lists instead
    of trees, but it can also have unintended consequences if a sequence is
    used in multiple parts of a grammar as the initial element of another
    sequence. Use a :py:class:`Wrapper` to prevent that from happening.

        .. code-block :: python

            a = Char("a")     # parses a single "a"
            b = Char("b")     # parses a single "b"
            c = Char("c")     # parses a single "c"

            ab = a + b        # parses a single "a" followed by a single "b"
                              # (a + b) creates a "Sequence" object. Using `ab`
                              # as an element in a later sequence would modify
                              # its original definition.

            abc = a + b + c   # parses "abc"
                              # (a + b) creates a "Sequence" object to which c
                              # is appended

            val = ab("ab")    # produces a list ["a", "b"]
            val = ab("a")     # raises an exception
            val = ab("b")     # raises an exception
            val = ab("ac")    # raises an exception
            val = ab("cb")    # raises an exception

            val = abc("abc")  # produces ["a", "b", "c"]
    """
    def __init__(self, children):
        super(Sequence, self).__init__()
        self.set_children(children)

    def __add__(self, other):
        return self.add_child(other)

    def process(self, pos, data, ctx):
        results = []
        for p in self.children:
            pos, res = p.process(pos, data, ctx)
            results.append(res)
        return pos, results


class Choice(Parser):
    """
    A Choice requires at least one of its children to succeed, and it returns
    the value of the one that matched. Alternatives in a choice are tried left
    to right, so they have a definite priority. This a feature of PEGs over
    context free grammars.

    Additional uses of ``|`` on the parser will cause it to accumulate parsers
    onto itself instead of creating new Choices. This has the desirable effect
    of increasing efficiency, but it can also have unintended consequences if a
    choice is used in multiple parts of a grammar as the initial element of
    another choice. Use a :py:class:`Wrapper` to prevent that from happening.

        .. code-block:: python

            abc = a | b | c   # alternation or choice.
            val = abc("a")    # parses a single "a"
            val = abc("b")    # parses a single "b"
            val = abc("c")    # parses a single "c"
            val = abc("d")    # raises an exception
    """
    def __init__(self, children):
        super(Choice, self).__init__()
        self.set_children(children)

    def __or__(self, other):
        return self.add_child(other)

    def process(self, pos, data, ctx):
        for c in self.children:
            try:
                return c.process(pos, data, ctx)
            except:
                pass
        raise Exception()


class Many(Parser):
    """
    Many wraps another parser and requires it to match a certain number of
    times.

    When Many matches zero occurences (``lower=0``), it always succeeds. Keep
    this in mind when using it in a list of alternatives or with
    :py:class:`FollowedBy` or :py:class:`NotFollowedBy`.

    The results are returned as a list.

        .. code-block:: python

            x = Char("x")
            xs = Many(x)      # parses many (or no) x's in a row
            val = xs("")      # returns []
            val = xs("a")     # returns []
            val = xs("x")     # returns ["x"]
            val = xs("xxxxx") # returns ["x", "x", "x", "x", "x"]
            val = xs("xxxxb") # returns ["x", "x", "x", "x"]

            ab = Many(a + b)  # parses "abab..."
            val = ab("")      # produces []
            val = ab("ab")    # produces [["a", b"]]
            val = ab("ba")    # produces []
            val = ab("ababab")# produces [["a", b"], ["a", "b"], ["a", "b"]]

            ab = Many(a | b)  # parses any combination of "a" and "b" like
                              # "aababbaba..."
            val = ab("aababb")# produces ["a", "a", "b", "a", "b", "b"]
            bs = Many(Char("b"), lower=1) # requires at least one "b"

    """
    def __init__(self, parser, lower=0):
        super(Many, self).__init__()
        self.add_child(parser)
        self.lower = lower

    def process(self, pos, data, ctx):
        orig = pos
        results = []
        p = self.children[0]
        while True:
            try:
                pos, res = p.process(pos, data, ctx)
                results.append(res)
            except Exception:
                break
        if len(results) < self.lower:
            child = self.children[0]
            msg = "Expected at least {0} of {1}.".format(self.lower, child)
            ctx.set(orig, msg)
            raise Exception()

        return pos, results

    def __repr__(self):
        if not self.name:
            return "Many({0}, lower={1})".format(self.children[0], self.lower)
        return super(Many, self).__repr__()


class Until(Parser):
    """
    Until wraps a parser and a terminal parser. It accumulates matches of the
    first parser until the terminal parser succeeds. Input for the terminal
    parser is left unread, and the results of the first parser are returned as
    a list.

    Since Until can match zero occurences, it always succeeds. Keep this in
    mind when using it in a list of alternatives or with :py:class:`FollowedBy`
    or :py:class:`NotFollowedBy`.

        .. code-block:: python

            cs = AnyChar.until(Char("y")) # parses many (or no) characters
                                          # until a "y" is encountered.

            val = cs("")                  # returns []
            val = cs("a")                 # returns ["a"]
            val = cs("x")                 # returns ["x"]
            val = cs("ccccc")             # returns ["c", "c", "c", "c", "c"]
            val = cs("abcdycc")           # returns ["a", "b", "c", "d"]

    """
    def __init__(self, parser, predicate):
        super(Until, self).__init__()
        self.set_children([parser, predicate])

    def process(self, pos, data, ctx):
        parser, pred = self.children
        results = []
        while True:
            try:
                pred.process(pos, data, ctx)
            except Exception:
                try:
                    pos, res = parser.process(pos, data, ctx)
                    results.append(res)
                except Exception:
                    break
            else:
                break
        return pos, results


class FollowedBy(Parser):
    """
    FollowedBy takes a parser and a predicate parser. The initial parser
    matches only if the predicate matches the input after it. On success, input
    for the predicate is left unread, and the result of the first parser is
    returned.

        .. code-block:: python

            ab = Char("a") & Char("b") # matches an "a" followed by a "b", but
                                       # the "b" isn't consumed from the input.
            val = ab("ab")             # returns "a" and leaves "b" to be
                                       # consumed.
            val = ab("ac")             # raises an exception and doesn't
                                       # consume "a".

    """
    def __init__(self, child, follow):
        super(FollowedBy, self).__init__()
        self.set_children([child, follow])

    def process(self, pos, data, ctx):
        left, right = self.children
        new, res = left.process(pos, data, ctx)
        right.process(new, data, ctx)
        return new, res


class NotFollowedBy(Parser):
    """
    NotFollowedBy takes a parser and a predicate parser. The initial parser
    matches only if the predicate parser fails to match the input after it. On
    success, input for the predicate is left unread, and the result of the
    first parser is returned.

        .. code-block:: python

            anb = Char("a") / Char("b") # matches an "a" not followed by a "b".
            val = anb("ac")             # returns "a" and leaves "c" to be
                                        # consumed
            val = anb("ab")             # raises an exception and doesn't
                                        # consume "a".

    """
    def __init__(self, child, follow):
        super(NotFollowedBy, self).__init__()
        self.set_children([child, follow])

    def process(self, pos, data, ctx):
        left, right = self.children
        new, res = left.process(pos, data, ctx)
        try:
            right.process(new, data, ctx)
        except Exception:
            return new, res
        else:
            msg = "{0} can't follow {1}".format(right, left)
            ctx.set(new, msg)
            raise Exception()


class KeepLeft(Parser):
    """
    KeepLeft takes two parsers. It requires them both to succeed but only
    returns results for the first one. It consumes input for both.

        .. code-block:: python

            a = Char("a")
            q = Char('"')

            aq = a << q      # like a + q except only the result of a is
                             # returned
            val = aq('a"')   # returns "a". Keeps the thing on the left of the
                             # <<

    """
    def __init__(self, left, right):
        super(KeepLeft, self).__init__()
        self.set_children([left, right])

    def process(self, pos, data, ctx):
        left, right = self.children
        pos, res = left.process(pos, data, ctx)
        pos, _ = right.process(pos, data, ctx)
        return pos, res


class KeepRight(Parser):
    """
    KeepRight takes two parsers. It requires them both to succeed but only
    returns results for the second one. It consumes input for both.

        .. code-block:: python

            q = Char('"')
            a = Char("a")

            qa = q >> a      # like q + a except only the result of a is
                             # returned
            val = qa('"a')   # returns "a". Keeps the thing on the right of the
                             # >>

    """
    def __init__(self, left, right):
        super(KeepRight, self).__init__()
        self.set_children([left, right])

    def process(self, pos, data, ctx):
        left, right = self.children
        pos, _ = left.process(pos, data, ctx)
        return right.process(pos, data, ctx)


class Opt(Parser):
    """
    Opt wraps a single parser and returns its value if it succeeds. It returns
    a default value otherwise. The input pointer is advanced only if the
    wrapped parser succeeds.

        .. code-block:: python

            a = Char("a")
            o = Opt(a)      # matches an "a" if its available. Still succeeds
                            # otherwise but doesn't advance the read pointer.
            val = o("a")    # returns "a"
            val = o("b")    # returns None. Read pointer is not advanced.

            o = Opt(a, default="x") # matches an "a" if its available. Returns
                                    # "x" otherwise.
            val = o("a")    # returns "a"
            val = o("b")    # returns "x". Read pointer is not advanced.

    """
    def __init__(self, p, default=None):
        super(Opt, self).__init__()
        self.add_child(p)
        self.default = default

    def process(self, pos, data, ctx):
        try:
            return self.children[0].process(pos, data, ctx)
        except Exception:
            return pos, self.default


class Map(Parser):
    """
    Map wraps a parser and a function. It returns the result of using the
    function to transform the wrapped parser's result.

    Example::

        .. code-block:: python

            Digit = InSet("0123456789")
            Digits = Many(Digit, lower=1)
            Number = Digits.map(lambda x: int("".join(x)))

    """
    def __init__(self, child, func):
        super(Map, self).__init__()
        self.add_child(child)
        self.func = func

    def process(self, pos, data, ctx):
        pos, res = self.children[0].process(pos, data, ctx)
        try:
            return pos, self.func(res)
        except Backtrack as bt:
            ctx.set(pos, bt.msg)
            raise
        except:
            tb = traceback.format_exc()
            msg = (self.name or "Map") + " raised{l}{tb}".format(l=os.linesep, tb=tb)
            ctx.function_error = (pos, msg)
            raise

    def __repr__(self):
        if not self.name:
            return "Map({0}({1}))".format(self.func.__name__, self.children[0])
        return super(Map, self).__repr__()


class Lift(Parser):
    """
    Lift wraps a function of multiple arguments. Use it with the multiplication
    operator on as many parsers as function arguments, and the results of those
    parsers will be passed to the function. The result of a Lift parser is the
    result of the wrapped function.

    Example::

        .. code-block:: python

            def comb(a, b, c):
                return "".join([a, b, c])

            # You'd normally invoke comb like comb("x", "y", "z"), but you can
            # "lift" it for use with parsers like this:

            x = Char("x")
            y = Char("y")
            z = Char("z")
            p = Lift(comb) * x * y * z

            # The * operator separates parsers whose results will go into the
            # arguments of the lifted function. I've used Char above, but x, y,
            # and z can be arbitrarily complex.

            val = p("xyz")  # would return "xyz"
            val = p("xyx")  # raises an exception. nothing would be consumed

    """
    def __init__(self, func):
        super(Lift, self).__init__()
        self.func = func

    def __mul__(self, other):
        return self.add_child(other)

    def process(self, pos, data, ctx):
        results = []
        for c in self.children:
            pos, res = c.process(pos, data, ctx)
            results.append(res)
        try:
            return pos, self.func(*results)
        except Backtrack as bt:
            ctx.set(pos, bt.msg)
            raise
        except:
            tb = traceback.format_exc()
            msg = (self.name or "Lift") + " raised{l}{tb}".format(l=os.linesep, tb=tb)
            ctx.set(pos, msg)
            ctx.function_error = (pos, msg)
            raise


class Forward(Parser):
    """
    Forward allows recursive grammars where a nonterminal's definition includes
    itself directly or indirectly. You initially create a Forward nonterminal
    with regular assignment.

        .. code-block:: python

            expr = Forward()

    You later give it its real definition with the ``<=`` operator.

        .. code-block:: python

            expr <= (term + Many(LowOps + term)).map(op)

    """
    def __init__(self):
        super(Forward, self).__init__()
        self.delegate = None

    def __le__(self, delegate):
        self.set_children([delegate])

    def process(self, pos, data, ctx):
        return self.children[0].process(pos, data, ctx)


class EOF(Parser):
    """
    EOF marks the end of input. This parser doesn't need to be created
    directly. An instance is provided in this module.

        .. code-block:: python

            # Top executes an Expr and ensures no input is left over.
            Top = Expr << EOF

    """
    def process(self, pos, data, ctx):
        if data[pos] is None:
            return pos, None
        msg = "Expected end of input."
        ctx.set(pos, msg)
        raise Exception(msg)


class EnclosedComment(Parser):
    """
    EnclosedComment matches a start literal, an end literal, and all characters
    between. It returns the content between the start and end.

        .. code-block:: python

            Comment = EnclosedComment("/*", "*/")

    """
    def __init__(self, s, e):
        super(EnclosedComment, self).__init__()
        Start = Literal(s)
        End = Literal(e)
        p = Start >> AnyChar.until(End).map(lambda x: "".join(x)) << End
        self.add_child(p)

    def process(self, pos, data, ctx):
        return self.children[0].process(pos, data, ctx)


class OneLineComment(Parser):
    """
    OneLineComment matches everything from a literal to the end of a line,
    excluding the end of line characters themselves. It returns the content
    between the start literal and the end of the line.

        .. code-block:: python

            Comment = OneLineComment("#") | OneLineComment("//")

    """
    def __init__(self, s):
        super(OneLineComment, self).__init__()
        p = Literal(s) >> Opt(AnyChar.until(InSet("\r\n")), "")
        self.add_child(p)

    def process(self, pos, data, ctx):
        return self.children[0].process(pos, data, ctx)


class WithIndent(Wrapper):
    """
    Consumes whitespace until a non-whitespace character is encountered, pushes
    the column position onto an indentation stack in the :py:class:`Context`,
    and then calls the parser it's wrapping. The wrapped parser and any
    of its children can make use of the saved indentation. Returns the value of
    the wrapped parser.

    WithIndent allows :py:class:`HangingString` to work by giving a way to mark
    how indented following lines must be to count as continuations.

        .. code-block:: python

            Key = WS >> PosMarker(String(key_chars)) << WS
            Sep = InSet(sep_chars, "Sep")
            Value = WS >> (Boolean | HangingString(value_chars))
            KVPair = WithIndent(Key + Opt(Sep >> Value))

    """
    def process(self, pos, data, ctx):
        new, _ = WS.process(pos, data, ctx)
        try:
            ctx.indents.append(ctx.col(new))
            return self.children[0].process(new, data, ctx)
        finally:
            ctx.indents.pop()


class HangingString(Parser):
    """
    HangingString matches lines with indented continuations like in ini files.

        .. code-block:: python

            Key = WS >> PosMarker(String(key_chars)) << WS
            Sep = InSet(sep_chars, "Sep")
            Value = WS >> (Boolean | HangingString(value_chars))
            KVPair = WithIndent(Key + Opt(Sep >> Value))

    """
    def __init__(self, chars, echars=None, min_length=1):
        super(HangingString, self).__init__()
        p = String(chars, echars=echars, min_length=min_length)
        self.add_child(p << (EOL | EOF))

    def process(self, pos, data, ctx):
        old = pos
        results = []
        while True:
            try:
                if ctx.col(pos) > ctx.indents[-1]:
                    pos, res = self.children[0].process(pos, data, ctx)
                    # Remove any inline comments.
                    results.append(res.split("#", 1)[0].rstrip(" \\"))
                else:
                    pos = old
                    break
                old = pos
                pos, _ = WS.process(pos, data, ctx)
            except Exception:
                break
        ret = " ".join(results)
        return pos, ret


class StartTagName(Wrapper):
    """
    Wraps a parser that represents a starting tag for grammars like xml, html,
    etc. The tag result is captured and put onto a tag stack in the
    :py:class:`Context` object.
    """
    def process(self, pos, data, ctx):
        pos, res = self.children[0].process(pos, data, ctx)
        ctx.tags.append(res)
        return pos, res


class EndTagName(Wrapper):
    """
    Wraps a parser that represents an end tag for grammars like xml, html, etc.
    The result is captured and compared to the last tag on the tag stack in the
    :py:class:`Context` object. The tags must match for the parse to be
    successful.
    """
    def __init__(self, parser, ignore_case=False):
        super(EndTagName, self).__init__(parser)
        self.ignore_case = ignore_case

    def process(self, pos, data, ctx):
        pos, res = self.children[0].process(pos, data, ctx)
        expect = ctx.tags.pop()

        r, e = res, expect

        if self.ignore_case:
            r = res.lower()
            e = expect.lower()

        if r != e:
            msg = "Expected {0!r}. Got {1!r}.".format(expect, res)
            ctx.set(pos, msg)
            raise Exception(msg)
        return pos, res


class EmptyQuotedString(Parser):
    def __init__(self, chars):
        super(EmptyQuotedString, self).__init__()
        single = Char("'") >> String(set(chars) - set("'"), "'", 0) << Char("'")
        double = Char('"') >> String(set(chars) - set('"'), '"', 0) << Char('"')
        self.add_child(single | double)

    def process(self, pos, data, ctx):
        return self.children[0].process(pos, data, ctx)


def _make_number(sign, int_part, frac_part):
    tmp = sign + int_part + ("".join(frac_part) if frac_part else "")
    return float(tmp) if "." in tmp else int(tmp)


def skip_none(x):
    return [i for i in x if i is not None]


EOF = EOF() % "EOF"
EOL = InSet("\n\r") % "EOL"
LineEnd = Wrapper(EOL | EOF) % "LineEnd"
EQ = Char("=")
LT = Char("<")
GT = Char(">")
FS = Char("/")
LeftCurly = Char("{")
RightCurly = Char("}")
LeftBracket = Char("[")
RightBracket = Char("]")
LeftParen = Char("(")
RightParen = Char(")")
Colon = Char(":")
SemiColon = Char(";")
Comma = Char(",")
AnyChar = AnyChar() % "any character"
NonZeroDigit = InSet(set(string.digits) - set("0")) % "non zero digit"
Digit = InSet(string.digits) % "digit"
Digits = String(string.digits) % "digits"
Letter = InSet(string.ascii_letters) % "ASCII letter"
Letters = String(string.ascii_letters) % "ASCII letters"
WSChar = InSet(set(string.whitespace) - set("\n\r")) % "whitespace w/o EOL"
WS = Many(InSet(string.whitespace) % "any whitespace")
Number = (Lift(_make_number) * Opt(Char("-"), "") * Digits * Opt(Char(".") + Digits)) % "number"
SingleQuotedString = Char("'") >> String(set(string.printable) - set("'"), "'") << Char("'")
DoubleQuotedString = Char('"') >> String(set(string.printable) - set('"'), '"') << Char('"')
QuotedString = Wrapper(DoubleQuotedString | SingleQuotedString) % "quoted string"



from insights.parsers.sysconfig import IrqbalanceSysconfig
from insights.tests import context_wrap

IRQBALANCE_SYSCONF_TEST = """
# irqbalance is a daemon process that distributes interrupts across
# CPUS on SMP systems. The default is to rebalance once every 10
# seconds. This is the environment file that is specified to systemd via the
# EnvironmentFile key in the service unit file (or via whatever method the init
# system you're using has.
#
# ONESHOT=yes
# after starting, wait for a minute, then look at the interrupt
# load and balance it once; after balancing exit and do not change
# it again.
#IRQBALANCE_ONESHOT=yes

#
# IRQBALANCE_BANNED_CPUS
# 64 bit bitmask which allows you to indicate which cpu's should
# be skipped when reblancing irqs. Cpu numbers which have their
# corresponding bits set to one in this mask will not have any
# irq's assigned to them on rebalance
#
IRQBALANCE_BANNED_CPUS=f8

#
# IRQBALANCE_ARGS
# append any args here to the irqbalance daemon as documented in the man page
#
IRQBALANCE_ARGS="-d"
""".strip()


def test_irqbalance_conf():
    ret = IrqbalanceSysconfig(context_wrap(IRQBALANCE_SYSCONF_TEST))
    assert ret['IRQBALANCE_BANNED_CPUS'] == 'f8'
    assert 'IRQBALANCE_ARGS' in ret
    assert ret.get('IRQBALANCE_ARGS') == '-d'
    assert 'IRQBALANCE_ONESHOT' not in ret



"""
Simultaneous Multithreading (SMT) parsers
=========================================

Parsers included in this module are:

CpuSMTActive - file ``/sys/devices/system/cpu/smt/active``
----------------------------------------------------------
CpuCoreOnline - files matching ``/sys/devices/system/cpu/cpu[0-9]*/online``
---------------------------------------------------------------------------
CpuSiblings - files matching ``/sys/devices/system/cpu/cpu[0-9]*/topology/thread_siblings_list``
------------------------------------------------------------------------------------------------
"""

from insights.core import Parser
from insights.core.plugins import parser
from insights.parsers import SkipException
from insights.specs import Specs
import re


@parser(Specs.cpu_smt_active)
class CpuSMTActive(Parser):
    """
    Class for parsing ``/sys/devices/system/cpu/smt/active`` file.
    Reports whether SMT is enabled and active.

    Typical output of this command is::

        1

    Raises:
        SkipException: When content is empty or cannot be parsed.

    Examples:
        >>> cpu_smt.on
        True
    """
    def parse_content(self, content):
        if not content:
            raise SkipException("No content.")
        self.on = bool(int(content[0]))


@parser(Specs.cpu_cores)
class CpuCoreOnline(Parser):
    """
    Class for parsing ``/sys/devices/system/cpu/cpu[0-9]*/online`` matching files.
    Reports whether a CPU core is online. Cpu0 is always online, so it does not have the "online" file.

    Typical output of this command is::

        1
        1
        1

    Raises:
        SkipException: When content is empty or cannot be parsed

    Examples:
        >>> cpu_core.core_id
        0
        >>> cpu_core.on
        True
    """
    cpu_core_path = r'/sys/devices/system/cpu/cpu(\d+)/online'

    def parse_content(self, content):
        if not content:
            raise SkipException("No content.")
        self.on = bool(int(content[0]))
        self.core_id = int(re.match(CpuCoreOnline.cpu_core_path, self.file_path).group(1))

    def __repr__(self):
        return "[Core {0}: {1}]".format(self.core_id, "Online" if self.on else "Offline")


@parser(Specs.cpu_siblings)
class CpuSiblings(Parser):
    """
    Class for parsing ``/sys/devices/system/cpu/cpu[0-9]*/topology/thread_siblings_list`` matching files.
    Reports CPU core siblings.

    Typical output of this command is::

        0,2
        1,3
        0,2
        1,3

    Raises:
        SkipException: When content is empty or cannot be parsed

    Examples:
        >>> cpu_siblings.core_id
        0
        >>> cpu_siblings.siblings
        [0, 2]
    """
    cpu_siblings_path = r'/sys/devices/system/cpu/cpu(\d+)/topology/thread_siblings_list'

    def parse_content(self, content):
        if not content:
            raise SkipException("No content.")

        # The separator in the sibling list may be either in the format 0-1 or 0,2 depending on the CPU model
        if "-" in content[0]:
            cpu_range = [int(x) for x in content[0].split("-")]
            self.siblings = [x for x in range(cpu_range[0], cpu_range[1] + 1)]
        elif "," in content[0]:
            self.siblings = [int(x) for x in content[0].split(",")]
        else:
            self.siblings = [int(content[0])]

        self.core_id = int(re.match(CpuSiblings.cpu_siblings_path, self.file_path).group(1))

    def __repr__(self):
        return "[Core {0} Siblings: {1}]".format(self.core_id, self.siblings)



"""
ProcKeys - File ``/proc/keys``
==============================

This parser reads the content of ``/proc/keys``.

"""

from insights import Parser, parser
from insights.parsers import SkipException, keyword_search
from insights.specs import Specs


@parser(Specs.proc_keys)
class ProcKeys(Parser, list):
    """
    Class ``ProcKeys`` parses the content of the ``/proc/keys`` file.

    This file exposes a list of the keys for which the reading thread has view
    permission, providing various information about each key. The fields shown
    in each line of this file contains below attributes::

        ID (string):            The ID of the key, expressed in hexadecimal.
        Flags (string):         A set of flags describing the state of the key
        Usage (string):         The count of the number of kernel credential structures
                                that are pinning the key.
        Timeout (string):       The amount of time until the key will expire (weeks, days,
                                hours, minutes, and seconds). The string perm here means
                                that the key is permanent (no timeout). The string expd
                                means that the key has already expired, but has not yet
                                been garbage collected.
        Permissions (string):   The key permissions, expressed as four hexadecimal bytes
                                containing, from left to right, the possessor, user, group,
                                and other permissions.
        UID (string):           The user ID of the key owner.
        GID (string):           The group ID of the key.
        Type (string) :         The key type.
        Description (string)    The key description (name). For most key types, it has the
                                form name[: extra-info]. (The name subfield is the key's
                                description (name). The optional extra-info field provides
                                some further information about the key.)


    Sample output::

        009a2028 I--Q---   1 perm 3f010000  1000  1000 user     krb_ccache:primary: 12
        1806c4ba I--Q---   1 perm 3f010000  1000  1000 keyring  _pid: 2
        25d3a08f I--Q---   1 perm 1f3f0000  1000 65534 keyring  _uid_ses.1000: 1
        28576bd8 I--Q---   3 perm 3f010000  1000  1000 keyring  _krb: 1
        2c546d21 I--Q--- 190 perm 3f030000  1000  1000 keyring  _ses: 2
        30a4e0be I------   4   2d 1f030000  1000 65534 keyring  _persistent.1000: 1
        32100fab I--Q---   4 perm 1f3f0000  1000 65534 keyring  _uid.1000: 2
        32a387ea I--Q---   1 perm 3f010000  1000  1000 keyring  _pid: 2
        3ce56aea I--Q---   5 perm 3f030000  1000  1000 keyring  _ses: 1


    Examples:
        >>> type(proc_keys)
        <class 'insights.parsers.proc_keys.ProcKeys'>
        >>> proc_keys[0]['id']
        '009a2028'
        >>> proc_keys[0]['flags']
        'I--Q---'
        >>> proc_keys[0]['usage']
        '1'
        >>> proc_keys[0]['timeout']
        'perm'
        >>> proc_keys[0]['permissions']
        '3f010000'
        >>> proc_keys[0]['uid']
        '1000'
        >>> proc_keys[0]['gid']
        '1000'
        >>> proc_keys[0]['type']
        'user'
        >>> proc_keys[0]['description']
        'krb_ccache:primary: 12'
    """

    def parse_content(self, content):

        if not content:
            raise SkipException("No Contents")

        column = ['id', 'flags', 'usage', 'timeout', 'permissions', 'uid', 'gid', 'type', 'description']

        for line in content:
            row = line.split(None, 8)
            if row and len(column) == len(row):
                self.append(dict(zip(column, row)))
            else:
                raise SkipException("Invalid Contents: {0}".format(line))

    def search(self, **kwargs):
        """
        Get the sublist containing the keywords by searching the ``/proc/keys`` list.

        This uses the :py:func:`insights.parsers.keyword_search` function for searching,
        see its documentation for usage details. If no search parameters are given or does
        match the search, then nothing will be returned.

        Returns:
            list: A list of dictionaries of the ``/proc/keys`` content that match the given search criteria.

        Examples:
            >>> proc_keys.search(timeout='perm')[0] == proc_keys[0]
            True
            >>> proc_keys.search(description__contains='uid')[0] == proc_keys[2]
            True
        """
        return keyword_search(self, **kwargs)



from .. import parser, get_active_lines, CommandParser
from . import parse_delimited_table
from insights.specs import Specs


def _replace_tabs(s, ts=8):
    """
    Replace the tabs in 's' and keep its original alignment with the tab-stop
    equals to 'ts'
    """
    result = ''
    for c in s:
        if c == '\t':
            while True:
                result += ' '
                if len(result) % ts == 0:
                    break
        else:
            result += c
    return result


def _lower(s, ic=True):
    return s.lower() if ic else s


class Table(object):

    def __init__(self, data):
        self.data = data


class TableSummary(Table):

    def __contains__(self, s):
        return any(s in r.get('table', '').lower() for r in self.data)

    def get(self, s, ic):
        s = _lower(s, ic)
        return [r for r in self.data if s in _lower(r.get('table', ''), ic)]


class ConstraintTable(Table):

    def __contains__(self, s):
        return any(s in r.get('table', r.get('TABLE NAME', '')).lower() for r in self.data)

    def get(self, s, ic):
        s = _lower(s, ic)
        return [r for r in self.data if s in _lower(r.get('table', r.get('TABLE NAME', '')), ic)]


class LabelTable(Table):
    pass


@parser(Specs.rhn_schema_stats)
class DBStatsLog(CommandParser):
    """
    Returns a DBStatsLog object which provides below two methods:
    - in:
        Returns if specified string (ignore case) is contained in any table
        name
    - get_table(s, ic=True):
        Returns the table info in which the table name contains specified
        string (ignore case by default)


    === Sample of PostgreSQL ==================================================
     schema |             table              |   rows
    --------+--------------------------------+----------
     public | rhnsnapshotpackage             | 47428950
     ...
    (402 rows)

     name                | type | table              | src
    ---------------------+------+--------------------+--------------
     chkpb_probe_type_ck | c    | rhn_check_probe    | ((probe_type))
     chkpb_probe_type    | c    | log                |
     ...
    (1402 rows)

     label |  created   |  modified  |       name       | epoch | version  | release
    -------+------------+------------+------------------+-------+----------+---------
    schema | 2015-10-16 | 2015-10-16 | satellite-schema |       | 5.7.0.19 | 1.el6sat
     ...
    (11 rows)

    === Sample of ORACLE ======================================================
    PLAN_TABLE: 0
    PXTSESSIONS: 10
    QRTZ_BLOB_TRIGGERS: 0
    ...

    CONSTRAINT NAME   TYPE TABLE NAME   SEARCH CONDITION    REFERENCED CONSTRAINT
    ----------------- ---- ------------ ------------------- ---------------------
    SYS_C003030       C    PXTSESSIONS  "VALUE" IS NOT NULL
    PXT_SESSIONS_PK   P    PXTSESSIONS
    \t\t\t...'Product E
    PXTSESSIONS_USER  R    PXTSESSIONS                      WEB_CONTACT_PK
    ...

    LABEL   CREATED    MODIFIED   NAME             EPOCH  VERSION    RELEASE
    ------- ---------- ---------- ---------------- ------ ---------- --------
    schema  2014-05-01 2014-05-01 satellite-schema        5.5.0.22   1.el5sat
    ==========================================================================
    """

    def __contains__(self, s):
        """
        Return if the 's' (ignore case) is contained in any table name or not
        """
        s = s.lower()
        return s in self.data.get("table_summary", []) or s in self.data.get("constraint_table", [])

    def get_table(self, s, ic=True):
        """
        Return the table info in which the table name contains
        the 's'
        """
        result = self.data["table_summary"].get(s, ic) if "table_summary" in self.data else []
        result.extend(self.data["constraint_table"].get(s, ic) if "constraint_table" in self.data else [])
        return result

    def parse_content(self, content):
        tables = []
        table = []
        if content and content[0].strip().startswith("schema"):
            # for PostgreSQL db stats log
            for line in get_active_lines(content, comment_char="--"):
                if line.startswith("(") and "rows" in line:
                    tables.append(parse_delimited_table(table, delim="|"))
                    table = []
                else:
                    table.append(line)
        else:
            # for Oracle db stats log
            header_line = []
            header_list = []
            index_list = []
            for line in content:
                if line.strip() == '' or line.startswith('\t'):
                    pass
                elif ': ' in line:
                    k, _, v = line.partition(': ')
                    table.append({'table': k.strip(),
                                  'rows': int(v.strip())})
                elif line.startswith('LABEL') or line.startswith('CONSTRAINT '):
                    header_line = _replace_tabs(line)
                elif header_line and line.startswith('--'):
                    pre_idx = 0
                    for idx, c in enumerate(line):
                        if c == ' ':
                            header_list.append(header_line[pre_idx:idx].strip())
                            pre_idx = idx + 1
                            index_list.append(pre_idx)
                    header_list.append(header_line[pre_idx:].strip())
                    index_list.append(-1)
                    header_line = []
                elif line.startswith("PL/SQL procedure ") or " rows " in line:
                    tables.append(table)
                    table = []
                    index_list = []
                    header_list = []
                elif len(index_list) > 1:
                    line = _replace_tabs(line)
                    items_list = []
                    pre_idx = 0
                    for idx in index_list:
                        items_list.append(line[pre_idx:idx].strip())
                        pre_idx = idx
                    table.append(dict(zip(header_list, items_list)))
        if tables:
            self.data = {
                "table_summary": TableSummary(tables[0]) if len(tables) > 0 else None,
                "constraint_table": ConstraintTable(tables[1]) if len(tables) > 1 else None,
                "label_table": LabelTable(tables[2]) if len(tables) > 2 else None
            }
        else:
            self.data = {}



from insights.tests import context_wrap
from insights.parsers.sysconfig import NfsSysconfig

SYSCONFIG_NFS = """
# Optional arguments passed to rpc.nfsd. See rpc.nfsd(8)
RPCNFSDARGS="--rdma=20049"
# Number of nfs server processes to be started.
# The default is 8.
#RPCNFSDCOUNT=16
#
# Set V4 grace period in seconds
#NFSD_V4_GRACE=90
#
# Set V4 lease period in seconds
#NFSD_V4_LEASE=90
#
# Optional arguments passed to rpc.mountd. See rpc.mountd(8)
RPCMOUNTDOPTS=""
# Port rpc.mountd should listen on.
#MOUNTD_PORT=892
#
# Optional arguments passed to rpc.statd. See rpc.statd(8)
STATDARG=""
# Port rpc.statd should listen on.
#STATD_PORT=662
# Outgoing port statd should used. The default is port
# is random
#STATD_OUTGOING_PORT=2020
# Specify callout program
#STATD_HA_CALLOUT="/usr/local/bin/foo"
#
#
# Optional arguments passed to sm-notify. See sm-notify(8)
SMNOTIFYARGS=""
#
# Optional arguments passed to rpc.idmapd. See rpc.idmapd(8)
RPCIDMAPDARGS=""
#
# Optional arguments passed to rpc.gssd. See rpc.gssd(8)
# Note: The rpc-gssd service will not start unless the
#       file /etc/krb5.keytab exists. If an alternate
#       keytab is needed, that separate keytab file
#       location may be defined in the rpc-gssd.service's
#       systemd unit file under the ConditionPathExists
#       parameter
RPCGSSDARGS=""
#
# Enable usage of gssproxy. See gssproxy-mech(8).
GSS_USE_PROXY="yes"
#
# Optional arguments passed to rpc.svcgssd. See rpc.svcgssd(8)
RPCSVCGSSDARGS=""
#
# Optional arguments passed to blkmapd. See blkmapd(8)
BLKMAPDARGS=""
RPCNFSDCOUNT=256
""".strip()


def test_sysconfig_nfs():
    result = NfsSysconfig(context_wrap(SYSCONFIG_NFS))
    assert result['GSS_USE_PROXY'] == 'yes'
    assert result['RPCNFSDARGS'] == '--rdma=20049'
    assert result.get('STATD_PORT') is None



from insights.parsr import Char


def test_and():
    p = Char("a") + Char("b")
    assert p("ab") == ["a", "b"]


def test_or():
    p = Char("a") | Char("b")
    assert p("a") == "a"
    assert p("b") == "b"



"""
Get SSL Certificate Info
========================

This module contains the following parsers:

SatelliteCustomCaChain - command ``awk 'BEGIN { pipe="openssl x509 -noout -subject -enddate"} /^-+BEGIN CERT/,/^-+END CERT/ { print | pipe } /^-+END CERT/ { close(pipe); printf("\\n")}' /etc/pki/katello/certs/katello-server-ca.crt``
========================================================================================================================================================================================================================================
RhsmKatelloDefaultCACert - command ``openssl x509 -in /etc/rhsm/ca/katello-default-ca.pem -noout -issuer``
==========================================================================================================
HttpdSSLCertExpireDate - command ``openssl x509 -in httpd_certificate_path -enddate -noout``
============================================================================================
NginxSSLCertExpireDate - command ``openssl x509 -in nginx_certificate_path -enddate -noout``
============================================================================================
MssqlTLSCertExpireDate - command ``openssl x509 -in mssql_tls_cert_file -enddate -noout``
============================================================================================
HttpdCertInfoInNSS - command ``certutil -L -d xxx -n xxx``
==========================================================
"""


from insights import parser, CommandParser
from datetime import datetime
from insights.core.dr import SkipComponent
from insights.parsers import ParseException, SkipException
from insights.specs import Specs
from insights.parsers.certificates_enddate import CertificatesEnddate


def parse_openssl_output(content):
    """
    It parses the output of "openssl -in <single_certificate_file> -xxx".
    Currently it only supports the attributes which the output is in
    key=value pairs. It saves the cert info into a dict. The value of notBefore and
    notAfter are saved to an instance of ExpirationDate, which contains the date
    in string and datetime format.

    Raises:
        ParseException: when the output isn't in key=value format or
                        the notAfter or notBefore isn't expected format.
    """
    date_format = '%b %d %H:%M:%S %Y'
    data = {}
    for line in content:
        if '=' not in line:
            raise ParseException('The line %s is not in key=value format' % line)
        key, value = [item.strip() for item in line.split('=', 1)]
        if key in ['notBefore', 'notAfter']:
            value_without_tz = value.rsplit(" ", 1)[0]
            try:
                date_time = datetime.strptime(value_without_tz, date_format)
            except Exception:
                raise ParseException('The %s is not in %s format.' % (key, date_format))
            value = CertificatesEnddate.ExpirationDate(value_without_tz, date_time)
        data[key] = value
    return data


class CertificateInfo(CommandParser, dict):
    """
    Base class to parse the output of "openssl -in <single_certificate_file> -xxx".
    Currently it only supports the attributes which the output is in
    key=value pairs.

    Sample Output::

        issuer= /C=US/ST=North Carolina/L=Raleigh/O=Katello/OU=SomeOrgUnit/CN=a.b.c.com
        notBefore=Dec  7 07:02:33 2020 GMT
        notAfter=Jan 18 07:02:33 2038 GMT
        subject= /C=US/ST=North Carolina/L=Raleigh/O=Katello/OU=SomeOrgUnit/CN=a.b.c.com

    Examples:
        >>> type(cert)
        <class 'insights.parsers.ssl_certificate.CertificateInfo'>
        >>> 'issuer' in cert
        True
        >>> cert['issuer']
        '/C=US/ST=North Carolina/L=Raleigh/O=Katello/OU=SomeOrgUnit/CN=a.b.c.com'
        >>> cert['notBefore'].str
        'Dec  7 07:02:33 2020'

    Raises:
        SkipException: when the command output is empty.
    """

    def __init__(self, context):
        super(CertificateInfo, self).__init__(
            context,
            extra_bad_lines=['error opening certificate', 'unable to load certificate'])

    def parse_content(self, content):
        """
        This uses the :py:func:`insights.parsers.ssl_certificate.parse_openssl_output` function.
        See its documentation for parsing details.
        """

        self.update(parse_openssl_output(content))
        if not self:
            raise SkipException("There is not any info in the cert.")

    @property
    def cert_path(self):
        '''Return the certificate path.'''
        return self.args


class CertificateChain(CommandParser, list):
    """
    Base class to parse the output of "openssl -in <certificate_chain_file> -xxx".
    Blank line is added to distinguish different certs in the chain.
    Currently it only supports the attributes which the output is in
    key=value pairs.

    Sample Output::

        issuer= /C=US/ST=North Carolina/L=Raleigh/O=Katello/OU=SomeOrgUnit/CN=test.a.com
        subject= /C=US/ST=North Carolina/L=Raleigh/O=Katello/OU=SomeOrgUnit/CN=test.b.com
        notBefore=Dec  7 07:02:33 2020 GMT
        notAfter=Jan 18 07:02:33 2038 GMT

        issuer= /C=US/ST=North Carolina/L=Raleigh/O=Katello/OU=SomeOrgUnit/CN=test.c.com
        subject= /C=US/ST=North Carolina/O=Katello/OU=SomeOrgUnit/CN=test.d.com
        notBefore=Nov 30 07:02:42 2020 GMT
        notAfter=Jan 18 07:02:43 2018 GMT

    Examples:
        >>> type(certs)
        <class 'insights.parsers.ssl_certificate.CertificateChain'>
        >>> len(certs)
        2
        >>> certs.earliest_expiry_date.str
        'Jan 18 07:02:43 2018'
    """

    def parse_content(self, content):
        """
        Parse the content of cert chain file. And it saves the certs
        in a list of dict.
        This uses the :py:func:`insights.parsers.ssl_certificate.parse_openssl_output` function.
        See its documentation for parsing details.

        Attributes:
            earliest_expiry_date(ExpirationDate):
                The earliest expiry datetime of the certs in the chain.
                None when there isn't "notAfter" for all the certs
                in the chain.

        Raises:
            SkipException: when the command output is empty.
        """

        self.earliest_expiry_date = None
        start_index = 0
        for index, line in enumerate(content):
            if not line.strip():
                # one cert ends
                if start_index != index:
                    self.append(parse_openssl_output(content[start_index:index]))
                start_index = index + 1
            if index == len(content) - 1:
                self.append(parse_openssl_output(content=content[start_index:index + 1]))
        if not self:
            raise SkipException("There is not any info in the ca cert chain.")
        for one_cert in self:
            expire_date = one_cert.get('notAfter')
            if expire_date and (self.earliest_expiry_date is None or expire_date.datetime < self.earliest_expiry_date.datetime):
                self.earliest_expiry_date = expire_date


@parser(Specs.satellite_custom_ca_chain)
class SatelliteCustomCaChain(CertificateChain):
    """
    .. note::
        Please refer to its super-class :class:`insights.parsers.ssl_certificate.CertificateChain` for more
        details.

    Sample Output::

        subject= /C=US/ST=North Carolina/L=Raleigh/O=Katello/OU=SomeOrgUnit/CN=test.a.com
        notAfter=Jan 18 07:02:33 2038 GMT

        subject= /C=US/ST=North Carolina/O=Katello/OU=SomeOrgUnit/CN=test.b.com
        notAfter=Jan 18 07:02:43 2028 GMT

    Examples:
        >>> type(satellite_ca_certs)
        <class 'insights.parsers.ssl_certificate.SatelliteCustomCaChain'>
        >>> len(satellite_ca_certs)
        2
        >>> satellite_ca_certs.earliest_expiry_date.str
        'Jan 18 07:02:43 2028'
    """
    pass


@parser(Specs.rhsm_katello_default_ca_cert)
class RhsmKatelloDefaultCACert(CertificateInfo):
    """
    .. note::
        Please refer to its super-class :class:`insights.parsers.ssl_certificate.CertificateInfo` for more
        details.

    Sample Output::

        issuer= /C=US/ST=North Carolina/L=Raleigh/O=Katello/OU=SomeOrgUnit/CN=a.b.c.com

    Examples:
        >>> type(rhsm_katello_default_ca)
        <class 'insights.parsers.ssl_certificate.RhsmKatelloDefaultCACert'>
        >>> rhsm_katello_default_ca['issuer']
        '/C=US/ST=North Carolina/L=Raleigh/O=Katello/OU=SomeOrgUnit/CN=a.b.c.com'
    """
    pass


@parser(Specs.httpd_ssl_cert_enddate)
class HttpdSSLCertExpireDate(CertificateInfo):
    """
    .. note::
        Please refer to its super-class :class:`insights.parsers.ssl_certificate.CertificateInfo` for more
        details.

    It parses the output of ``openssl x509 -in httpd_ssl_certificate_path -enddate -noout``.

    Sample output of ``openssl x509 -in httpd_certificate_path -enddate -noout``::

        notAfter=Dec 4 07:04:05 2035 GMT

    Examples:
        >>> type(date_info)
        <class 'insights.parsers.ssl_certificate.HttpdSSLCertExpireDate'>
        >>> date_info['notAfter'].datetime
        datetime.datetime(2038, 1, 18, 7, 2, 43)
    """
    pass


@parser(Specs.nginx_ssl_cert_enddate)
class NginxSSLCertExpireDate(CertificateInfo):
    """
    .. note::
        Please refer to its super-class :class:`insights.parsers.ssl_certificate.CertificateInfo` for more
        details.

    It parses the output of ``openssl x509 -in nginx_certificate_path -enddate -noout``.

    Sample output of ``openssl x509 -in nginx_certificate_path -enddate -noout``::

        notAfter=Dec 4 07:04:05 2035 GMT

    Examples:
        >>> type(nginx_date_info)
        <class 'insights.parsers.ssl_certificate.NginxSSLCertExpireDate'>
        >>> nginx_date_info['notAfter'].datetime
        datetime.datetime(2038, 1, 18, 7, 2, 43)
        >>> nginx_date_info.cert_path
        '/a/b/c.pem'
    """
    pass


@parser(Specs.mssql_tls_cert_enddate)
class MssqlTLSCertExpireDate(CertificateInfo):
    """
    .. note::
        Please refer to its super-class :class:`insights.parsers.ssl_certificate.CertificateInfo` for more
        details.

    It parses the output of ``openssl x509 -in mssql_tls_cert_file -enddate -noout``.

    Sample output of ``openssl x509 -in mssql_tls_cert_file -enddate -noout``::

        notAfter=Dec 4 07:04:05 2035 GMT

    Examples:
        >>> type(mssql_date_info)
        <class 'insights.parsers.ssl_certificate.MssqlTLSCertExpireDate'>
        >>> mssql_date_info['notAfter'].datetime
        datetime.datetime(2022, 11, 5, 1, 43, 59)
    """
    pass


@parser(Specs.httpd_cert_info_in_nss)
class HttpdCertInfoInNSS(CommandParser, dict):
    """
    It parses the output of "certutil -d <database_path> -L -n <cert_name>".
    Currently it only parses the "Not After" info and save it into a dict.
    And the key is renamed to "notAfter" to keep consistent with the other certificat info.
    The value of "notAfter" is transformed to an instance of ExpirationDate,
    which contains the date in string and datetime format.

    Raises:
        ParseException: when the "Not After" isn't in the expected format.
        SkipComponent: when there is no "Not After" info in the content.

    Examples:
        >>> type(nss_cert_info)
        <class 'insights.parsers.ssl_certificate.HttpdCertInfoInNSS'>
        >>> nss_cert_info['notAfter'].str
        'Sun Dec 07 05:26:10 2025'
    """
    date_format = '%a %b %d %H:%M:%S %Y'

    def parse_content(self, content):
        # currently only expire date is needed
        for line in content:
            if 'Not After :' in line:
                key, value = [item.strip() for item in line.split(':', 1)]
                try:
                    date_time = datetime.strptime(value, self.date_format)
                except Exception:
                    raise ParseException('The %s is not in %s format.' % (key, self.date_format))
                value = CertificatesEnddate.ExpirationDate(value, date_time)
                self.update({'notAfter': value})
        if not self:
            raise SkipComponent

    @property
    def cert_path(self):
        '''Return the certificate path info.'''
        return self.args



from insights import dr


class needs(dr.ComponentType):
    group = "needs"


@needs()
def one():
    return 1


@needs()
def two():
    return 2


@needs(one, two)
def report(a, b):
    return a + b


def test_single_component():
    graph = dr.get_dependency_graph(report)
    components = dr._determine_components(report)
    assert graph == components


def test_list_of_components():
    graph = dr.get_dependency_graph(report)
    components = dr._determine_components([one, two, report])
    assert graph == components


def test_group():
    graph = dr.get_dependency_graph(report)
    components = dr._determine_components("needs")
    assert graph == components


def test_type():
    graph = dr.get_dependency_graph(report)
    components = dr._determine_components(needs)
    assert graph == components



"""
PulpWorkerDefaults - file ``/etc/default/pulp_workers``
=======================================================

The PulpWorkerDefaults parser reads the shell options set in the
``/etc/default/pulp_worker`` file.  These are made available using the
:class:`insights.core.SysconfigOptions` class methods.

Sample file contents::

    # Configuration file for Pulp's Celery workers

    # Define the number of worker nodes you wish to have here. This defaults to the number of processors
    # that are detected on the system if left commented here.
    PULP_CONCURRENCY=1

    # Configure Python's encoding for writing all logs, stdout and stderr
    PYTHONIOENCODING="UTF-8"

Examples:
    >>> pulp_defs = shared[PulpWorkerDefaults]
    >>> type(pulp_defs)
    <class 'insights.parsers.pulp_worker_defaults.PulpWorkerDefaults'>
    >>> 'PULP_CONCURRENCY' in pulp_defs
    True
    >>> pulp_defs['PULP_CONCURRENCY']  # Note string return value
    '1'
    >>> 'PULP_MAX_TASKS_PER_CHILD' in pulp_defs
    False
    >>> pulp_defs['PYTHONIOENCODING']  # Values are dequoted as per bash
    'UTF-8'

"""

from .. import parser, SysconfigOptions
from insights.specs import Specs


@parser(Specs.pulp_worker_defaults)
class PulpWorkerDefaults(SysconfigOptions):
    """
    Parse the ``/etc/default/pulp_workers`` file.
    """
    pass



import doctest

from insights.parsers import redhat_release
from insights.parsers.redhat_release import RedhatRelease
from insights.tests import context_wrap


REDHAT_RELEASE1 = """
Red Hat Enterprise Linux Server release 6.7 (Santiago)
""".strip()

REDHAT_RELEASE2 = """
Red Hat Enterprise Linux Server release 7.2 (Maipo)
""".strip()

REDHAT_RELEASE3 = """
Red Hat Enterprise Linux release 7.5-0.14
""".strip()

RHVH_RHV40 = """
Red Hat Enterprise Linux release 7.3
""".strip()

RHEVH_RHEV35 = """
Red Hat Enterprise Virtualization Hypervisor release 6.7 (20160219.0.el6ev)
""".strip()

FEDORA = """
Fedora release 23 (Twenty Three)
""".strip()

REDHAT_RELEASE8 = """
Red Hat Enterprise Linux release 8.2 (Ootpa)
""".strip()

REDHAT_RELEASE10 = """
Red Hat Enterprise Linux Server release 6.10(Santiago)
""".strip()

REDHAT_RELEASE_BETA = """
Red Hat Enterprise Linux Server release 8.5 Beta (Ootpa)
""".strip()

CENTOS_STREAM = """
CentOS Stream release 8
""".strip()

CENTOS_7 = """
CentOS Linux release 7.6.1810 (Core)
""".strip()

REDHAT_RELEASE_ALPHA = """
Red Hat Enterprise Linux release 9.0 Alpha (Plow)
""".strip()


def test_rhe6():
    release = RedhatRelease(context_wrap(REDHAT_RELEASE1))
    assert release.raw == REDHAT_RELEASE1
    assert release.major == 6
    assert release.minor == 7
    assert release.version == "6.7"
    assert release.is_rhel
    assert release.product == "Red Hat Enterprise Linux Server"


def test_rhe7():
    release = RedhatRelease(context_wrap(REDHAT_RELEASE2))
    assert release.raw == REDHAT_RELEASE2
    assert release.major == 7
    assert release.minor == 2
    assert release.version == "7.2"
    assert release.is_rhel
    assert release.product == "Red Hat Enterprise Linux Server"


def test_rhe75_0_14():
    release = RedhatRelease(context_wrap(REDHAT_RELEASE3))
    assert release.raw == REDHAT_RELEASE3
    assert release.major == 7
    assert release.minor == 5
    assert release.version == "7.5-0.14"
    assert release.is_rhel
    assert release.product == "Red Hat Enterprise Linux"


def test_rhevh35():
    release = RedhatRelease(context_wrap(RHEVH_RHEV35))
    assert release.raw == RHEVH_RHEV35
    assert release.major == 6
    assert release.minor == 7
    assert release.version == "6.7"
    assert not release.is_rhel
    assert release.product == "Red Hat Enterprise Virtualization Hypervisor"


def test_rhvh40():
    release = RedhatRelease(context_wrap(RHVH_RHV40))
    assert release.raw == RHVH_RHV40
    assert release.major == 7
    assert release.minor == 3
    assert release.version == "7.3"
    assert release.is_rhel
    assert release.product == "Red Hat Enterprise Linux"


def test_fedora23():
    release = RedhatRelease(context_wrap(FEDORA))
    assert release.raw == FEDORA
    assert release.major == 23
    assert release.minor is None
    assert release.version == "23"
    assert not release.is_rhel
    assert release.is_fedora
    assert release.product == "Fedora"


def test_rhel6_10():
    release = RedhatRelease(context_wrap(REDHAT_RELEASE10))
    assert release.raw == REDHAT_RELEASE10
    assert release.major == 6
    assert release.minor == 10
    assert release.version == "6.10"
    assert release.is_rhel
    assert release.product == "Red Hat Enterprise Linux Server"


def test_rhel8():
    release = RedhatRelease(context_wrap(REDHAT_RELEASE8))
    assert release.raw == REDHAT_RELEASE8
    assert release.major == 8
    assert release.minor == 2
    assert release.version == "8.2"
    assert release.is_rhel
    assert release.product == "Red Hat Enterprise Linux"


def test_rhel_alpha():
    release = RedhatRelease(context_wrap(REDHAT_RELEASE_ALPHA))
    assert release.raw == REDHAT_RELEASE_ALPHA
    assert release.major == 9
    assert release.minor == 0
    assert release.version == "9.0"
    assert release.is_rhel
    assert release.is_alpha
    assert not release.is_beta
    assert release.parsed['code_name'] == 'Plow'
    assert release.product == "Red Hat Enterprise Linux"


def test_rhel_beta():
    release = RedhatRelease(context_wrap(REDHAT_RELEASE_BETA))
    assert release.raw == REDHAT_RELEASE_BETA
    assert release.major == 8
    assert release.minor == 5
    assert release.version == "8.5"
    assert release.is_rhel
    assert release.is_beta
    assert not release.is_alpha
    assert release.parsed['code_name'] == 'Ootpa'
    assert release.product == "Red Hat Enterprise Linux Server"


def test_centos_stream():
    release = RedhatRelease(context_wrap(CENTOS_STREAM))
    assert release.major == 8
    assert release.minor is None
    assert release.product == 'CentOS Stream'
    assert release.is_centos
    assert not release.is_rhel


def test_centos_7():
    release = RedhatRelease(context_wrap(CENTOS_7))
    assert release.major == 7
    assert release.minor == 6
    assert release.product == 'CentOS Linux'
    assert release.code_name == 'Core'
    assert release.is_centos
    assert not release.is_rhel


def test_examples():
    release = RedhatRelease(context_wrap(REDHAT_RELEASE2))
    globs = {
        'rh_release': release
    }
    failed, tested = doctest.testmod(redhat_release, globs=globs)
    assert failed == 0



import pytest
from insights.core.dr import SkipComponent
from insights.parsers import lsof
from insights.tests import context_wrap

LSOF = """
lsof: avoiding stat(/): -b was specified.
COMMAND     PID  TID           USER   FD      TYPE             DEVICE  SIZE/OFF       NODE NAME
systemd       1                root  cwd       DIR              253,1      4096        128 /
systemd       1                root  rtd       DIR              253,1      4096        128 /
systemd       1                root  txt       REG              253,1   1230920    1440410 /usr/lib/systemd/systemd
systemd       1                root  mem       REG              253,1     37152  135529970 /usr/lib64/libnss_sss.so.2
meventd     674  688           root  txt       REG              253,1     36344  135317954 /usr/sbin/dmeventd
dmeventd    674  688           root  mem       REG              253,1     20032  134413763 /usr/lib64/libuuid.so.1.3.0
dmeventd    674  688           root  mem       REG              253,1    248584  135108725 /usr/lib64/libblkid.so.1.1.0
bioset      611                root  txt   unknown                                         /proc/611/exe
systemd       1                root  cwd       DIR              253,1      4096        128 /
""".strip()

LSOF_GOOD_V1 = """
COMMAND    PID  TID           USER   FD      TYPE             DEVICE  SIZE/OFF       NODE NAME
systemd-l  602                root   14u      CHR              13,64       0t0       6406 /dev/input/event0
systemd-l  602                root   15u      CHR              13,65       0t0       6407 /dev/input/event1
systemd-l  602                root   16u      CHR                4,6       0t0       4694 /dev/tty6
dbus-daem  603                dbus    0r      CHR                1,3       0t0       4674 /dev/null
dbus-daem  603  615           dbus    0r      CHR                1,3       0t0       4674 /dev/null
abrt-watc 8619                root    0r      CHR                1,3       0t0       4674 /dev/null
wpa_suppl  641                root    0r      CHR                1,3       0t0       4674 /dev/null
polkitd    642             polkitd    0u      CHR                1,3       0t0       4674 /dev/null
polkitd    642             polkitd    1u      CHR                1,3       0t0       4674 /dev/null
polkitd    642             polkitd    2u      CHR                1,3       0t0       4674 /dev/null
gmain      642  643        polkitd    0u      CHR                1,3       0t0       4674 /dev/null
gmain      642  643        polkitd    1u      CHR                1,3       0t0       4674 /dev/null
gmain      642  643        polkitd    2u      CHR                1,3       0t0       4674 /dev/null
gdbus      642  646        polkitd    0u      CHR                1,3       0t0       4674 /dev/null
gdbus      642  646        polkitd    1u      CHR                1,3       0t0       4674 /dev/null
gdbus      642  646        polkitd    2u      CHR                1,3       0t0       4674 /dev/null
JS         642  648        polkitd    1r      BLK               8,16       0t0      16884 /dev/sda
JS         642  648        polkitd    1u      CHR                1,3       0t0       4674 /dev/null
JS         642  648        polkitd    2u      CHR                1,3       0t0       4674 /dev/null
""".strip()

LSOF_GOOD_V2 = """
lsof: avoiding readlink(/sys): -b was specified.
lsof: avoiding stat(/sys): -b was specified.
lsof: WARNING: can't stat() sysfs file system /sys
      Output information may be incomplete.
COMMAND       PID     USER   FD      TYPE             DEVICE  SIZE/OFF       NODE NAME
systemd         1        0  cwd       DIR              253,0       251        128 /
systemd         1        0    0u      CHR                1,3       0t0       1032 /dev/null
systemd         1        0   19u     unix 0xffff998f20251f80       0t0  132266239 /run/systemd/journal/stdout type=STREAM
blkcg_pun      36        0  txt   unknown                                         /proc/36/exe
rsyslogd   772018        0    5w      REG              253,6 151175680       1351 /var/log/messages-20210221 (deleted)
""".strip()

columns = ["COMMAND", "PID", "TID", "USER", "FD", "TYPE", "DEVICE", "SIZE/OFF", "NODE", "NAME"]


def test_lsof():
    ctx = context_wrap(LSOF)
    d = list(lsof.Lsof(ctx).parse(LSOF.splitlines()))

    assert set(columns) == set([k for f in d for k in f.keys()])
    assert d[0] == {
        "COMMAND": "systemd",
        "PID": "1",
        "TID": "",
        "USER": "root",
        "FD": "cwd",
        "TYPE": "DIR",
        "DEVICE": "253,1",
        "SIZE/OFF": "4096",
        "NODE": "128",
        "NAME": "/"
    }

    # Spot checks
    assert d[3]["TID"] == ""
    assert d[4]["TID"] == "688"
    assert d[5]["TYPE"] == "REG"
    assert d[7]["NAME"] == "/proc/611/exe"

    assert d[-2] == {
        "COMMAND": "bioset",
        "PID": "611",
        "TID": "",
        "USER": "root",
        "FD": "txt",
        "TYPE": "unknown",
        "DEVICE": "",
        "SIZE/OFF": "",
        "NODE": "",
        "NAME": "/proc/611/exe"
    }
    assert d[-1] == {
        "COMMAND": "systemd",
        "PID": "1",
        "TID": "",
        "USER": "root",
        "FD": "cwd",
        "TYPE": "DIR",
        "DEVICE": "253,1",
        "SIZE/OFF": "4096",
        "NODE": "128",
        "NAME": "/"
    }


def test_lsof_good():
    ctx = context_wrap(LSOF_GOOD_V1)
    d = list(lsof.Lsof(ctx).parse(LSOF_GOOD_V1.splitlines()))
    assert d[0] == {
        "COMMAND": "systemd-l",
        "PID": "602",
        "TID": "",
        "USER": "root",
        "FD": "14u",
        "TYPE": "CHR",
        "DEVICE": "13,64",
        "SIZE/OFF": "0t0",
        "NODE": "6406",
        "NAME": "/dev/input/event0"
    }

    # Spot checks
    assert d[3]["TID"] == ""
    assert d[4]["TID"] == "615"
    assert d[5]["COMMAND"] == "abrt-watc"
    assert d[5]["PID"] == "8619"
    assert d[5]["TYPE"] == "CHR"
    assert d[7]["NAME"] == "/dev/null"

    assert d[-1] == {
        "COMMAND": "JS",
        "PID": "642",
        "TID": "648",
        "USER": "polkitd",
        "FD": "2u",
        "TYPE": "CHR",
        "DEVICE": "1,3",
        "SIZE/OFF": "0t0",
        "NODE": "4674",
        "NAME": "/dev/null"
    }


def test_lsof_good_v2():
    ctx = context_wrap(LSOF_GOOD_V2)
    d = list(lsof.Lsof(ctx).parse(LSOF_GOOD_V2.splitlines()))
    assert d[0] == {
        'COMMAND': 'systemd',
        'PID': '1',
        'USER': '0',
        'FD': 'cwd',
        'TYPE': 'DIR',
        'DEVICE': '253,0',
        'SIZE/OFF': '251',
        'NODE': '128',
        'NAME': '/'
    }

    # wide and empty values checks
    assert d[2]['DEVICE'] == '0xffff998f20251f80'
    assert d[3]['DEVICE'] == ''

    assert d[-1] == {
        'COMMAND': 'rsyslogd',
        'PID': '772018',
        'USER': '0',
        'FD': '5w',
        'TYPE': 'REG',
        'DEVICE': '253,6',
        'SIZE/OFF': '151175680',
        'NODE': '1351',
        'NAME': '/var/log/messages-20210221 (deleted)'
    }


def test_lsof_scan():
    ctx = context_wrap(LSOF_GOOD_V1)
    # Scannable provided `any` method
    lsof.Lsof.any('systemd_commands', lambda x: 'systemd' in x['COMMAND'])
    # Scannable provided `collect` method
    lsof.Lsof.collect('polkitd_user', lambda x: x['USER'] == 'polkitd')
    # Lsof provided `collect_keys` method
    lsof.Lsof.collect_keys('root_stdin', USER='root', FD='0r', SIZE_OFF='0t0')
    l = lsof.Lsof(ctx)
    assert l.systemd_commands
    assert len(l.polkitd_user) == 12

    assert hasattr(l, 'root_stdin')
    assert len(l.root_stdin) == 2

    assert l.root_stdin[0]['COMMAND'] == 'abrt-watc'
    assert l.root_stdin[0]['PID'] == '8619'
    assert l.root_stdin[0]['TID'] == ''
    assert l.root_stdin[0]['USER'] == 'root'
    assert l.root_stdin[0]['FD'] == '0r'
    assert l.root_stdin[0]['TYPE'] == 'CHR'
    assert l.root_stdin[0]['DEVICE'] == '1,3'
    assert l.root_stdin[0]['SIZE/OFF'] == '0t0'
    assert l.root_stdin[0]['NODE'] == '4674'
    assert l.root_stdin[0]['NAME'] == '/dev/null'

    assert l.root_stdin[1]['COMMAND'] == 'wpa_suppl'
    assert l.root_stdin[1]['PID'] == '641'
    assert l.root_stdin[1]['TID'] == ''
    assert l.root_stdin[1]['USER'] == 'root'
    assert l.root_stdin[1]['FD'] == '0r'
    assert l.root_stdin[1]['TYPE'] == 'CHR'
    assert l.root_stdin[1]['DEVICE'] == '1,3'
    assert l.root_stdin[1]['SIZE/OFF'] == '0t0'
    assert l.root_stdin[1]['NODE'] == '4674'
    assert l.root_stdin[1]['NAME'] == '/dev/null'


LSOF_BAD = """
lsof: WARNING: can't stat() xfs file system /var/lib/origin/openshift.local.volumes/pods/abca2a21-da4c-22eb-b49c-011d3a938eb5/volume-subpaths/config/wh/0
"""


def test_lsof_bad():
    with pytest.raises(SkipComponent) as e:
        lsof.Lsof(context_wrap(LSOF_BAD))
    assert e is not None



"""
LogrotateConf - files ``/etc/logrotate.conf`` and others
========================================================

Class to parse logrotate confuration files:
- ``/etc/logrotate.conf``
- ``/etc/logrotate.d/*``

See: http://www.linuxmanpages.org/8/logrotate
"""
import string

from insights.core import ConfigParser, LegacyItemAccess, Parser
from insights.core.plugins import parser
from insights.parsers import ParseException, get_active_lines
from insights.parsr import (AnyChar, Choice, EOF, EOL, Forward, LeftCurly, LineEnd,
                            Literal, Many, Number, OneLineComment, Opt, PosMarker,
                            QuotedString, RightCurly, skip_none, String, WS, WSChar)
from insights.parsr.query import Directive, Entry, Section
from insights.specs import Specs

PAIRED_OPTS = (
        'prerotate',
        'postrotate',
        'firstaction',
        'lastaction',
)


@parser(Specs.logrotate_conf)
class LogrotateConf(Parser, LegacyItemAccess):
    """
    Class for parsing ``/etc/logrotate.conf`` and ``/etc/logrotate.d/*``
    configuration files.

    Sample logrotate configuration file::

        # sample file
        compress

        /var/log/messages {
            rotate 5
            weekly
            postrotate
                        /sbin/killall -HUP syslogd
            endscript
        }

        "/var/log/httpd/access.log" /var/log/httpd/error.log {
            rotate 5
            mail www@my.org
            size=100k
            sharedscripts
            postrotate
                        /sbin/killall -HUP httpd
            endscript
        }

        /var/log/news/news.crit
        /var/log/news/olds.crit  {
            monthly
            rotate 2
            olddir /var/log/news/old
            missingok
            postrotate
                        kill -HUP `cat /var/run/inn.pid`
            endscript
            nocompress
        }

    Examples:
        >>> type(log_rt)
        <class 'insights.parsers.logrotate_conf.LogrotateConf'>
        >>> log_rt.options
        ['compress']
        >>> log_rt.log_files
        ['/var/log/messages', '/var/log/httpd/access.log', '/var/log/httpd/error.log', '/var/log/news/news.crit', '/var/log/news/olds.crit']
        >>> log_rt['compress']
        True
        >>> 'weekly' in log_rt['/var/log/messages']
        True
        >>> log_rt['/var/log/messages']['postrotate']
        ['/sbin/killall -HUP syslogd']

    Attributes:
        data(dict): All parsed options and log files are stored in this
            dictionary
        options(list): List of global options in the configuration file
        log_files(list): List of log files in the configuration file
    """

    def parse_content(self, content):

        def _parse_opts(line):
            if '=' in line:
                l_sp = line.split('=', 1)
            else:
                l_sp = line.split(None, 1)
            # return a (key, value) tuple
            return l_sp[0], l_sp[1] if len(l_sp) == 2 else True

        self.data = {}
        self.options = []
        self.log_files = []

        log_opts = script = None
        log_files = []
        for line in get_active_lines(content):
            if log_opts is None:
                # not in log_file section
                if line.lstrip(' \t\'"').startswith('/'):
                    # log_file line
                    log_files.extend([l.strip(' \t\'"') for l in line.rstrip('{').split()])
                if line.endswith('{'):
                    # start of the section
                    log_opts = {}  # empty dict indicates in a log_file section
                elif not log_files:
                    # global options
                    key1, val = _parse_opts(line)
                    self.data[key1] = val
                    self.options.append(key1)
            else:
                # in log_file section
                if line.endswith('}') and script is None:
                    # end of the section,
                    # save options for each log_file individually
                    for lf in log_files:
                        self.data[lf] = log_opts
                    # collect all log_files
                    self.log_files.extend(log_files)
                    log_files = []
                    log_opts = None
                    continue
                # script options
                if line in PAIRED_OPTS:
                    # in script section
                    script = line
                    log_opts[script] = []
                    continue
                if line == 'endscript':
                    # end of script section
                    script = None
                    continue
                if script is not None:
                    # scripts inner PARED_OPTS
                    log_opts[script].append(line)
                else:
                    # common options in log_file section
                    key, val = _parse_opts(line)
                    log_opts[key] = val


class DocParser(object):
    def __init__(self, ctx):
        self.ctx = ctx

        scripts = set("postrotate prerotate firstaction lastaction preremove".split())
        Stanza = Forward()
        Spaces = Many(WSChar)
        Bare = String(set(string.printable) - (set(string.whitespace) | set("#{}'\"")))
        Num = Number & (WSChar | LineEnd)
        Comment = OneLineComment("#").map(lambda x: None)
        ScriptStart = WS >> PosMarker(Choice([Literal(s) for s in scripts])) << WS
        ScriptEnd = Literal("endscript")
        Line = (WS >> AnyChar.until(EOL) << WS).map(lambda x: "".join(x))
        Lines = Line.until(ScriptEnd | EOF).map(lambda x: "\n".join(x))
        Script = ScriptStart + Lines << Opt(ScriptEnd)
        Script = Script.map(lambda x: [x[0], [x[1]], None])
        BeginBlock = WS >> LeftCurly << WS
        EndBlock = WS >> RightCurly
        First = PosMarker((Bare | QuotedString)) << Spaces
        Attr = Spaces >> (Num | Bare | QuotedString) << Spaces
        Rest = Many(Attr)
        Block = BeginBlock >> Many(Stanza).map(skip_none).map(self.to_entries) << EndBlock
        Stmt = WS >> (Script | (First + Rest + Opt(Block))) << WS
        Stanza <= WS >> (Stmt | Comment) << WS
        Doc = Many(Stanza).map(skip_none).map(self.to_entries)

        self.Top = Doc << EOF

    def to_entries(self, x):
        ret = []
        for i in x:
            name, attrs, body = i
            if body:
                for n in [name.value] + attrs:
                    ret.append(Section(name=n, children=body, lineno=name.lineno))
            else:
                ret.append(Directive(name=name.value, attrs=attrs, lineno=name.lineno))
        return ret

    def __call__(self, content):
        try:
            return self.Top(content)
        except Exception:
            raise ParseException("There was an exception when parsing the logrotate config file.")


def parse_doc(content, ctx=None):
    """ Parse a configuration document into a tree that can be queried. """
    if isinstance(content, list):
        content = "\n".join(content)
    parse = DocParser(ctx)
    result = parse(content)
    return Entry(children=result, src=ctx)


@parser(Specs.logrotate_conf)
class LogRotateConfPEG(ConfigParser):
    """
    Sample logrotate configuration file::

        # sample file
        compress

        /var/log/messages {
            rotate 5
            weekly
            postrotate
                        /sbin/killall -HUP syslogd
            endscript
        }

        "/var/log/httpd/access.log" /var/log/httpd/error.log {
            rotate 5
            mail www@my.org
            size=100k
            sharedscripts
            postrotate
                        /sbin/killall -HUP httpd
            endscript
        }

        /var/log/news/news.crit
        /var/log/news/olds.crit  {
            monthly
            rotate 2
            olddir /var/log/news/old
            missingok
            postrotate
                        kill -HUP `cat /var/run/inn.pid`
            endscript
            nocompress
        }

    Examples:
        >>> type(log_rt_peg)
        <class 'insights.parsers.logrotate_conf.LogRotateConfPEG'>
        >>> 'compress' in log_rt_peg
        True
        >>> 'weekly' in log_rt['/var/log/messages']
        True
        >>> log_rt_peg['/var/log/messages']['postrotate'][-1].value
        '/sbin/killall -HUP syslogd'
    """
    def parse_doc(self, content):
        return parse_doc("\n".join(content), ctx=self)



import os
from setuptools import setup, find_packages

__here__ = os.path.dirname(os.path.abspath(__file__))

package_info = dict.fromkeys(["RELEASE", "COMMIT", "VERSION", "NAME"])

for name in package_info:
    with open(os.path.join(__here__, "insights", name)) as f:
        package_info[name] = f.read().strip()

entry_points = {
    'console_scripts': [
        'insights-collect = insights.collect:main',
        'insights-run = insights:main',
        'insights = insights.command_parser:main',
        'insights-cat = insights.tools.cat:main',
        'insights-dupkeycheck = insights.tools.dupkeycheck:main',
        'insights-inspect = insights.tools.insights_inspect:main',
        'insights-info = insights.tools.query:main',
        'insights-ocpshell= insights.ocpshell:main',
        'mangle = insights.util.mangle:main'
    ]
}

runtime = set([
    'six',
    'requests',
    'redis',
    'cachecontrol',
    'cachecontrol[redis]',
    'cachecontrol[filecache]',
    'defusedxml',
    'lockfile',
    'jinja2<=2.11.3; python_version <= "2.7"',
    'jinja2; python_version > "2.7"',
    'pyyaml>=3.10,<=3.13; python_version < "2.7"',
    'pyyaml; python_version >= "2.7"',
])


def maybe_require(pkg):
    try:
        __import__(pkg)
    except ImportError:
        runtime.add(pkg)


maybe_require("importlib")
maybe_require("argparse")


client = set([
    'requests',
    'python-gnupg==0.4.6',
    'oyaml'
])

develop = set([
    'wheel',
])

docs = set([
    'docutils',
    'Sphinx',
    'nbsphinx',
    'sphinx_rtd_theme',
    'ipython',
    'colorama',
    'Pygments',
    'jedi<0.18.0',    # Open issue with jedi 0.18.0 and iPython <= 7.19
                      # https://github.com/davidhalter/jedi/issues/1714
])

# python 2.6 requires setuptools~=36.8.0 to support this syntax
testing = set([
    'coverage==4.3.4; python_version < "2.7"',
    'coverage; python_version >= "2.7"',
    'pytest==3.0.6; python_version < "2.7"',
    'pytest~=4.6.0; python_version == "2.7"',
    'pytest; python_version >= "3"',
    'pytest-cov==2.4.0; python_version < "2.7"',
    'pytest-cov; python_version >= "2.7"',
    'mock==2.0.0',
])

cluster = set([
    'ansible',
    'pandas',
    'colorama',
])

openshift = set([
    'openshift'
])

linting = set([
    'flake8==2.6.2; python_version < "2.7"',
    'flake8; python_version >= "2.7"'
])

optional = set([
    'python-cjson',
    'python-logstash',
    'python-statsd',
    'watchdog',
])

if __name__ == "__main__":
    # allows for runtime modification of rpm name
    name = os.environ.get("INSIGHTS_CORE_NAME", package_info["NAME"])

    setup(
        name=name,
        version=package_info["VERSION"],
        description="Insights Core is a data collection and analysis framework",
        long_description=open("README.rst").read(),
        url="https://github.com/redhatinsights/insights-core",
        author="Red Hat, Inc.",
        author_email="insights@redhat.com",
        packages=find_packages(),
        install_requires=list(runtime),
        package_data={'': ['LICENSE']},
        license='Apache 2.0',
        extras_require={
            'develop': list(runtime | develop | client | docs | linting | testing | cluster),
            'develop26': list(runtime | develop | client | linting | testing | cluster),
            'client': list(runtime | client),
            'client-develop': list(runtime | develop | client | linting | testing),
            'cluster': list(runtime | cluster),
            'openshift': list(runtime | openshift),
            'optional': list(optional),
            'docs': list(docs),
            'linting': list(linting | client),
            'testing': list(testing | client)
        },
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Natural Language :: English',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
        ],
        entry_points=entry_points,
        include_package_data=True
    )
